@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
@@ -99,15 +99,20 @@ void csync_mark(const char *file, const char *thispeer, const char *peerfilter)
 
 	csync_debug(1, "Marking file as dirty: %s\n", file);
 	for (pl_idx=0; pl[pl_idx].peername; pl_idx++)
-		if (!peerfilter || !strcmp(peerfilter, pl[pl_idx].peername))
+		if (!peerfilter || !strcmp(peerfilter, pl[pl_idx].peername)) {
+			SQL("Deleting old dirty file entries",
+				"DELETE FROM dirty WHERE filename = '%s' AND peername = '%s'",
+				url_encode(file),
+				url_encode(pl[pl_idx].peername));
+
 			SQL("Marking File Dirty",
-				"%s INTO dirty (filename, force, myname, peername) "
+				"INSERT INTO dirty (filename, forced, myname, peername) "
 				"VALUES ('%s', %s, '%s', '%s')",
-				csync_new_force ? "REPLACE" : "INSERT",
 				url_encode(file),
 				csync_new_force ? "1" : "0",
 				url_encode(pl[pl_idx].myname),
 				url_encode(pl[pl_idx].peername));
+		}
 
 	free(pl);
 }
@@ -122,21 +127,87 @@ int csync_check_pure(const char *filename)
 	if (!csync_lowercyg_disable)
 		return 0;
 #endif
-
 	struct stat sbuf;
-	int i=0;
+	int dir_len = 0;
+	int i;
+	int same_len;
 
-	while (filename[i]) i++;
+	/* single entry last query cache
+	 * to speed up checks from deep subdirs */
+	static struct {
+		/* store inclusive trailing slash for prefix match */
+		char *path;
+		/* strlen(path) */
+		int len;
+		/* cached return value */
+		int has_symlink;
+	} cached;
+
+	for (i = 0; filename[i]; i++)
+		if (filename[i] == '/')
+			dir_len = i+1;
+
+	if (dir_len <= 1) /* '/' a symlink? hardly. */
+		return 0;
+
+	/* identical prefix part */
+	for (i = 0; i < dir_len && i < cached.len; i++)
+		if (filename[i] != cached.path[i])
+			break;
+
+	/* backtrack to slash */
+	for (--i; i >= 0 && cached.path[i] != '/'; --i);
+		;
+
+	same_len = i+1;
+
+	csync_debug(3, " check: %s %u, %s %u, %u.\n",
+		filename, dir_len,
+		cached.path ?: "(null)", cached.len,
+		same_len);
+
+	/* exact match? */
+	if (dir_len == same_len && same_len == cached.len)
+		return cached.has_symlink;
 
 	{ /* new block for myfilename[] */
-		char myfilename[i+1];
-		memcpy(myfilename, filename, i+1);
-		while (1) {
-			while (myfilename[i] != '/')
-				if (--i <= 0) return 0;
+		char myfilename[dir_len+1];
+		char *to_be_cached;
+		int has_symlink = 0;
+		memcpy(myfilename, filename, dir_len);
+		myfilename[dir_len] = '\0';
+		to_be_cached = strdup(myfilename);
+		i = dir_len-1;
+		while (i) {
+			for (; i && myfilename[i] != '/'; --i)
+				;
+
+			if (i <= 1)
+				break;
+
+			if (i+1 == same_len) {
+				if (same_len == cached.len) {
+					/* exact match */
+					has_symlink = cached.has_symlink;
+					break;
+				} else if (!cached.has_symlink)
+					/* prefix of something 'pure' */
+					break;
+			}
+
 			myfilename[i]=0;
-			if ( lstat_strict(prefixsubst(myfilename), &sbuf) || S_ISLNK(sbuf.st_mode) ) return 1;
+			if (lstat_strict(prefixsubst(myfilename), &sbuf) || S_ISLNK(sbuf.st_mode)) {
+				has_symlink = 1;
+				break;
+			}
 		}
+		if (to_be_cached) { /* strdup can fail. So what. */
+			free(cached.path);
+			cached.path = to_be_cached;
+			cached.len = dir_len;
+			cached.has_symlink = has_symlink;
+		}
+		return has_symlink;
 	}
 }
 
@@ -148,18 +219,22 @@ void csync_check_del(const char *file, int recursive, int init_run)
 
 	if ( recursive ) {
 		if ( !strcmp(file, "/") )
-			asprintf(&where_rec, "or 1");
+		  ASPRINTF(&where_rec, "OR 1=1");
 		else
-			asprintf(&where_rec, "or (filename > '%s/' "
-					"and filename < '%s0')",
-					url_encode(file), url_encode(file));
+		  ASPRINTF(&where_rec, "UNION ALL SELECT filename from file where filename > '%s/' "
+				"and filename < '%s0'",
+				url_encode(file), url_encode(file));
 	}
 
 	SQL_BEGIN("Checking for removed files",
 			"SELECT filename from file where "
 			"filename = '%s' %s ORDER BY filename", url_encode(file), where_rec)
 	{
-		const char *filename = url_decode(SQL_V[0]);
+		const char *filename = url_decode(SQL_V(0));
+
+		if (!csync_match_file(filename))
+			continue;
+
 		if ( lstat_strict(prefixsubst(filename), &st) != 0 || csync_check_pure(filename) )
 			textlist_add(&tl, filename, 0);
 	} SQL_END;
@@ -196,6 +271,7 @@ int csync_check_mod(const char *file, int recursive, int ignnoent, int init_run)
 			if (!strcmp(file, p->path)) {
 				char new_file[strlen(p->name) + 3];
 				sprintf(new_file, "%%%s%%", p->name);
+				csync_debug(2, "Prefix matched: %s <- %s.\n", new_file, file);
 				csync_check_mod(new_file, recursive, ignnoent, init_run);
 				continue;
 			}
@@ -231,7 +307,7 @@ int csync_check_mod(const char *file, int recursive, int ignnoent, int init_run)
 			"filename = '%s'", url_encode(file))
 		{
 			if ( !csync_cmpchecktxt(checktxt,
-						url_decode(SQL_V[0])) ) {
+						url_decode(SQL_V(0))) ) {
 				csync_debug(2, "File has changed: %s\n", file);
 				this_is_dirty = 1;
 			}
@@ -243,6 +319,10 @@ int csync_check_mod(const char *file, int recursive, int ignnoent, int init_run)
 		} SQL_END;
 
 		if ( this_is_dirty && !csync_compare_mode ) {
+			SQL("Deleting old file entry",
+			    "DELETE FROM file WHERE filename = '%s'",
+			    url_encode(file));
+
 			SQL("Adding or updating file entry",
 			    "INSERT INTO file (filename, checktxt) "
 			    "VALUES ('%s', '%s')",
