@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
@@ -36,18 +36,27 @@
 #include <errno.h>
 #include <signal.h>
 #include <ctype.h>
+#include <syslog.h>
+#include "db_api.h"
+#include <netdb.h>
 
 #ifdef REAL_DBDIR
 #  undef DBDIR
 #  define DBDIR REAL_DBDIR
 #endif
 
-static char *file_database = 0;
+char *csync_database = 0;
+
+int db_type = DB_SQLITE3;
+
 static char *file_config = 0;
-static char *dbdir = DBDIR;
 char *cfgname = "";
+char *systemdir = NULL;	/* ETCDIR */
+char *lockfile = NULL;	/* ETCDIR/csync2.lock */
 
 char myhostname[256] = "";
+int bind_to_myhostname = 0;
+char *csync_port = "30865";
 char *active_grouplist = 0;
 char *active_peerlist = 0;
 
@@ -57,11 +66,11 @@ extern FILE *yyin;
 int csync_error_count = 0;
 int csync_debug_level = 0;
 FILE *csync_debug_out = 0;
+int csync_syslog = 0;
 
 int csync_server_child_pid = 0;
 int csync_timestamps = 0;
 int csync_new_force = 0;
-int csync_port = 30865;
 
 int csync_dump_dir_fd = -1;
 
@@ -91,8 +100,12 @@ void help(char *cmd)
 	printf(
 "\n"
 PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
-"LINBIT Information Technologies GmbH <http://www.linbit.com>\n"
-"Copyright (C) 2004, 2005  Clifford Wolf <clifford@clifford.at>\n"
+"Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH\n"
+"        http://www.linbit.com\n"
+"See also: http://git.linbit.com/csync2.git/?a=blob;f=AUTHORS\n"
+"\n"
+"Version: " CSYNC2_VERSION "\n"
+"\n"
 "This program is free software under the terms of the GNU GPL.\n"
 "\n"
 "Usage: %s [-v..] [-C config-name] \\\n"
@@ -103,7 +116,7 @@ PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
 "	-c [-r] file..		Check files and maybe add to dirty db\n"
 "	-u [-d] [-r] file..	Updates files if listed in dirty db\n"
 "	-o [-r] file..		Create list of files in compare-mode\n"
-"	-f [-r] file..		Force this file in sync (resolve conflict)\n"
+"	-f [-r] file..		Force files to win next conflict resolution\n"
 "	-m file..		Mark files in database as dirty\n"
 "\n"
 "Simple mode:\n"
@@ -131,15 +144,18 @@ PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
 "\n"
 "	-TT	As -T, but print the unified diffs.\n"
 "\n"
-"	The modes -H, -L, -M and -S return 2 if the requested db is empty.\n"
-"	The mode -T returns 2 if both hosts are in sync.\n"
-"\n"
 "	-i	Run in inetd server mode.\n"
 "	-ii	Run in stand-alone server mode.\n"
 "	-iii	Run in stand-alone server mode (one connect only).\n"
 "\n"
 "	-R	Remove files from database which do not match config entries.\n"
 "\n"
+"Exit codes:\n"
+"	The modes -H, -L, -M and -S return 2 if the requested db is empty.\n"
+"	The mode -T returns 2 if both hosts are in sync.\n"
+"	Otherwise, only exit codes 0 (no errors)\n"
+"	and 1 (some unspecified errrors) are expected.\n"
+"\n"
 "Modifiers:\n"
 "	-r	Recursive operation over subdirectories\n"
 "	-d	Dry-run on all remote update operations\n"
@@ -161,11 +177,11 @@ PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
 "	-U	Don't mark all other peers as dirty when doing a -TI run.\n"
 "\n"
 "	-G Group1,Group2,Group3,...\n"
-"		Only use this groups from config-file.\n"
+"		Only use these groups from config-file.\n"
 "\n"
 "	-P peer1,peer1,...\n"
-"		Only update this peers (still mark all as dirty).\n"
-"		Only show files for this peers in -o (compare) mode.\n"
+"		Only update these peers (still mark all as dirty).\n"
+"		Only show files for these peers in -o (compare) mode.\n"
 "\n"
 "	-F	Add new entries to dirty database with force flag set.\n"
 "\n"
@@ -178,10 +194,30 @@ PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
 "		found to the specified file descriptor (when doing a -c run).\n"
 "		The directory names in this output are zero-terminated.\n"
 "\n"
+"Database switches:\n"
+"	-D database-dir or url\n"
+"		default: /var/lib/csync2\n"
+"		Absolute path: use sqlite database in that directory\n"
+"	    URLs:\n"
+"		sqlite:///some/path[/database.db3]\n"
+"		sqlite3:///some/path[/database.db3]\n"
+"		sqlite2:///some/path[/database.db]\n"
+"		mysql://[<user>:<password>@]<hostname>/[database]\n"
+"		pgsql://[<user>:<password>@]<hostname>/[database]\n"
+"	If database is not given, it defaults to csync2_<qualified hostname>.\n"
+"	Note that for non-sqlite backends, the database name is \"cleaned\",\n"
+"	characters outside of [0-9][a-z][A-Z] will be replaced with _.\n"
+"\n"
 "Creating key file:\n"
 "	%s -k filename\n"
 "\n"
-"Csync2 will refuse to do anything when a " ETCDIR "/csync2.lock file is found.\n"
+"Environment variables:\n"
+"	CSYNC2_SYSTEM_DIR\n"
+"		Directory containing csync2.cfg and other csync2 system files.\n"
+"		Defaults to " ETCDIR ".\n"
+"\n"
+"Csync2 will refuse to do anything if this file is found:\n"
+"$CSYNC2_SYSTEM_DIR/csync2.lock\n"
 "\n",
 		cmd, cmd);
 	exit(1);
@@ -190,70 +226,137 @@ PACKAGE_STRING " - cluster synchronization tool, 2nd generation\n"
 int create_keyfile(const char *filename)
 {
 	int fd = open(filename, O_WRONLY|O_CREAT|O_EXCL, 0600);
-	int rand = open("/dev/random", O_RDONLY);
+	int rand = open("/dev/urandom", O_RDONLY);
 	char matrix[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._";
 	unsigned char n;
 	int i;
-
+	int rc;
 	assert(sizeof(matrix) == 65);
 	if ( fd == -1 ) {
 		fprintf(stderr, "Can't create key file: %s\n", strerror(errno));
 		return 1;
 	}
 	if ( rand == -1 ) {
-		fprintf(stderr, "Can't open /dev/random: %s\n", strerror(errno));
+		fprintf(stderr, "Can't open /dev/urandom: %s\n", strerror(errno));
 		return 1;
 	}
 	for (i=0; i<64; i++) {
-		read(rand, &n, 1);
-		write(fd, matrix+(n&63), 1);
+		rc = read(rand, &n, 1);
+		rc = write(fd, matrix+(n&63), 1);
 	}
-	write(fd, "\n", 1);
+	rc = write(fd, "\n", 1);
 	close(rand);
 	close(fd);
 	return 0;
 }
 
-static int csync_server_loop(int single_connect)
+static int csync_server_bind(void)
 {
 	struct linger sl = { 1, 5 };
-	struct sockaddr_in addr;
-	int on = 1;
+	struct addrinfo hints;
+	struct addrinfo *result, *rp;
+	int save_errno;
+	int sfd = -1, s, on = 1;
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_family = AF_UNSPEC;	/* Allow IPv4 or IPv6 */
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_flags = AI_PASSIVE;
 
-	int listenfd = socket(AF_INET, SOCK_STREAM, 0);
+	s = getaddrinfo(bind_to_myhostname ? myhostname : NULL, csync_port, &hints, &result);
+	if (s != 0) {
+		csync_debug(1, "Cannot prepare local socket, getaddrinfo: %s\n", gai_strerror(s));
+		return -1;
+	}
+
+	/* getaddrinfo() returns a list of address structures.
+	   Try each address until we successfully bind(2).
+	   If socket(2) (or bind(2)) fails, we (close the socket
+	   and) try the next address. */
+
+	for (rp = result; rp != NULL; rp = rp->ai_next) {
+		sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
+		if (sfd == -1)
+			continue;
+
+		if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &on, (socklen_t) sizeof(on)) < 0)
+			goto error;
+		if (setsockopt(sfd, SOL_SOCKET, SO_LINGER, &sl, (socklen_t) sizeof(sl)) < 0)
+			goto error;
+		if (setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, &on, (socklen_t) sizeof(on)) < 0)
+			goto error;
+
+		if (bind(sfd, rp->ai_addr, rp->ai_addrlen) == 0)
+			break;	/* Success */
+
+		close(sfd);
+		sfd = -1;
+	}
+
+	if (sfd != -1) {
+		char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
+		if (getnameinfo(rp->ai_addr, rp->ai_addrlen,
+				hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
+				NI_NUMERICHOST | NI_NUMERICSERV) == 0)
+			csync_debug(1, "Listening on %s:%s as %s.\n",
+				hbuf, sbuf, myhostname);
+		else
+			/* WTF, is failure even possible here? */
+			csync_debug(1, "Listening on <?>:%s as %s.\n",
+				csync_port, myhostname);
+	}
+
+	freeaddrinfo(result);	/* No longer needed */
+
+	return sfd;
+
+error:
+	save_errno = errno;
+	close(sfd);
+	errno = save_errno;
+	return -1;
+}
+
+static int csync_server_loop(int single_connect)
+{
+	union {
+		struct sockaddr sa;
+		struct sockaddr_in sa_in;
+		struct sockaddr_in6 sa_in6;
+		struct sockaddr_storage ss;
+	} addr;
+	int listenfd = csync_server_bind();
 	if (listenfd < 0) goto error;
 
-	bzero(&addr, sizeof(addr));
-	addr.sin_family = AF_INET;
-	addr.sin_addr.s_addr = htonl(INADDR_ANY);
-	addr.sin_port = htons(csync_port);
-
-	if ( setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, (socklen_t) sizeof(on)) < 0 ) goto error;
-	if ( setsockopt(listenfd, SOL_SOCKET, SO_LINGER, &sl, (socklen_t) sizeof(sl)) < 0 ) goto error;
-	if ( setsockopt(listenfd, IPPROTO_TCP, TCP_NODELAY, &on, (socklen_t) sizeof(on)) < 0 ) goto error;
-
-	if ( bind(listenfd, (struct sockaddr *) &addr, sizeof(addr)) < 0 ) goto error;
-	if ( listen(listenfd, 5) < 0 ) goto error;
+	if (listen(listenfd, 5) < 0) goto error;
 
+	/* we want to "cleanly" shutdown if the connection is lost unexpectedly */
 	signal(SIGPIPE, SIG_IGN);
+	/* server is not interested in its childs, prevent zombies */
 	signal(SIGCHLD, SIG_IGN);
 
 	printf("Csync2 daemon running. Waiting for connections.\n");
 
 	while (1) {
-		int addrlen = sizeof(addr);
-		int conn = accept(listenfd, (struct sockaddr *) &addr, &addrlen);
+		unsigned addrlen = sizeof(addr);
+		int conn = accept(listenfd, &addr.sa, &addrlen);
 		if (conn < 0) goto error;
 
 		fflush(stdout); fflush(stderr);
 
+		if (single_connect)
+			close(listenfd);
+
 		if (single_connect || !fork()) {
+			char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
+			/* need to restore default SIGCHLD handler in the session,
+			 * as we may need to wait on them in action.c */
+			signal(SIGCHLD, SIG_DFL);
 			csync_server_child_pid = getpid();
-			fprintf(stderr, "<%d> New connection from %s:%u.\n",
-				csync_server_child_pid,
-				inet_ntoa(addr.sin_addr),
-				ntohs(addr.sin_port));
-			fflush(stderr);
+			if (getnameinfo(&addr.sa, addrlen,
+					hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
+					NI_NUMERICHOST | NI_NUMERICSERV) != 0)
+				goto error;
+			csync_debug(1, "New connection from %s:%s.\n", hbuf, sbuf);
 
 			dup2(conn, 0);
 			dup2(conn, 1);
@@ -265,10 +368,16 @@ static int csync_server_loop(int single_connect)
 	}
 
 error:
-	fprintf(stderr, "Server error: %s\n", strerror(errno));
+	csync_debug(0, "Server error: %s\n", strerror(errno));
 	return 1;
 }
 
+void csync_openlog(void)
+{
+	csync_syslog = 1;
+	openlog("csync2", LOG_ODELAY | LOG_PID, LOG_LOCAL0);
+}
+
 int main(int argc, char ** argv)
 {
 	struct textlist *tl = 0, *t;
@@ -288,12 +397,18 @@ int main(int argc, char ** argv)
 		return create_keyfile(argv[2]);
 	}
 
-	if (!access(ETCDIR "/csync2.lock", F_OK)) {
-		printf("Found " ETCDIR "/csync2.lock.\n");
+	systemdir = getenv("CSYNC2_SYSTEM_DIR");
+	if (!systemdir)
+		systemdir = ETCDIR;
+
+	ASPRINTF(&lockfile, "%s/csync2.lock", systemdir);
+	if (!access(lockfile , F_OK)) {
+		printf("Found %s\n", lockfile);
 		return 1;
 	}
 
-	while ( (opt = getopt(argc, argv, "W:s:Ftp:G:P:C:D:N:HBAIXULSTMRvhcuoimfxrd")) != -1 ) {
+	while ( (opt = getopt(argc, argv, "W:s:Ftp:G:P:C:D:N:HBAIXULlSTMRvhcuoimfxrd")) != -1 ) {
+
 		switch (opt) {
 			case 'W':
 				csync_dump_dir_fd = atoi(optarg);
@@ -314,7 +429,7 @@ int main(int argc, char ** argv)
 				csync_timestamps = 1;
 				break;
 			case 'p':
-				csync_port = atoi(optarg);
+				csync_port = strdup(optarg);
 				break;
 			case 'G':
 				active_grouplist = optarg;
@@ -341,14 +456,18 @@ int main(int argc, char ** argv)
 				cfgname = optarg;
 				break;
 			case 'D':
-				dbdir = optarg;
+				csync_database = optarg;
 				break;
 			case 'N':
 				snprintf(myhostname, 256, "%s", optarg);
+				++bind_to_myhostname;
 				break;
 			case 'v':
 				csync_debug_level++;
 				break;
+			case 'l':
+				csync_openlog();
+				break;
 			case 'h':
 				if ( mode != MODE_NONE ) help(argv[0]);
 				mode = MODE_HINT;
@@ -450,6 +569,11 @@ int main(int argc, char ** argv)
 	if ( mode == MODE_NONE )
 		help(argv[0]);
 
+	/* Some inetd connect stderr to stdout.  The debug level messages on
+	 * stderr would confuse the csync2 protocol. Log to syslog instead. */
+	if ( mode == MODE_INETD && csync_debug_level && !csync_syslog )
+		csync_openlog();
+
 	if ( *myhostname == 0 ) {
 		gethostname(myhostname, 256);
 		myhostname[255] = 0;
@@ -472,21 +596,22 @@ int main(int argc, char ** argv)
 	 * before we open the config file and database
 	 */
 	if ( mode == MODE_INETD ) {
-		char line[4096], *cmd, *para;
+		static char line[4 * 4096];
+		char *cmd, *para;
 
 		/* configure conn.c for inetd mode */
 		conn_set(0, 1);
 
-		if ( !conn_gets(line, 4096) ) return 0;
+		if ( !conn_gets(line, sizeof(line)) ) return 0;
 		cmd = strtok(line, "\t \r\n");
 		para = cmd ? strtok(0, "\t \r\n") : 0;
 
 		if (cmd && !strcasecmp(cmd, "ssl")) {
-#ifdef HAVE_LIBGNUTLS_OPENSSL
-			conn_printf("OK (activating_ssl).\n");
+#ifdef HAVE_LIBGNUTLS
+			conn_resp(CR_OK_ACTIVATING_SSL);
 			conn_activate_ssl(1);
 
-			if ( !conn_gets(line, 4096) ) return 0;
+			if ( !conn_gets(line, sizeof(line)) ) return 0;
 			cmd = strtok(line, "\t \r\n");
 			para = cmd ? strtok(0, "\t \r\n") : 0;
 #else
@@ -503,10 +628,8 @@ int main(int argc, char ** argv)
 		if (para)
 			cfgname = strdup(url_decode(para));
 	}
-
 	if ( !*cfgname ) {
-		asprintf(&file_database, "%s/%s.db", dbdir, myhostname);
-		asprintf(&file_config, ETCDIR "/csync2.cfg");
+	     ASPRINTF(&file_config, "%s/csync2.cfg", systemdir);
 	} else {
 		int i;
 
@@ -518,14 +641,10 @@ int main(int argc, char ** argv)
 				return mode != MODE_INETD;
 			}
 
-		asprintf(&file_database, "%s/%s_%s.db", dbdir, myhostname, cfgname);
-		asprintf(&file_config, ETCDIR "/csync2_%s.cfg", cfgname);
+		ASPRINTF(&file_config, "%s/csync2_%s.cfg", systemdir, cfgname);
 	}
 
-	csync_debug(2, "My hostname is %s.\n", myhostname);
-	csync_debug(2, "Database-File: %s\n", file_database);
 	csync_debug(2, "Config-File:   %s\n", file_config);
-
 	yyin = fopen(file_config, "r");
 	if ( !yyin )
 		csync_fatal("Can not open config file `%s': %s\n",
@@ -533,6 +652,12 @@ int main(int argc, char ** argv)
 	yyparse();
 	fclose(yyin);
 
+	if (!csync_database || !csync_database[0] || csync_database[0] == '/')
+		csync_database = db_default_database(csync_database);
+
+	csync_debug(2, "My hostname is %s.\n", myhostname);
+	csync_debug(2, "Database-File: %s\n", csync_database);
+
 	{
 		const struct csync_group *g;
 		for (g=csync_group; g; g=g->next)
@@ -541,7 +666,7 @@ int main(int argc, char ** argv)
 found_a_group:;
 	}
 
-	csync_db_open(file_database);
+	csync_db_open(csync_database);
 
 	for (i=optind; i < argc; i++)
 		on_cygwin_lowercase(argv[i]);
@@ -582,8 +707,8 @@ found_a_group:;
 				SQL_BEGIN("Check all hints",
 					"SELECT filename, recursive FROM hint")
 				{
-					textlist_add(&tl, url_decode(SQL_V[0]),
-							atoi(SQL_V[1]));
+					textlist_add(&tl, url_decode(SQL_V(0)),
+							atoi(SQL_V(1)));
 				} SQL_END;
 
 				for (t = tl; t != 0; t = t->next) {
@@ -635,58 +760,63 @@ found_a_group:;
 			break;
 
 		case MODE_INETD:
-			conn_printf("OK (cmd_finished).\n");
+			conn_resp(CR_OK_CMD_FINISHED);
 			csync_daemon_session();
 			break;
 
 		case MODE_MARK:
 			for (i=optind; i < argc; i++) {
 				char *realname = getrealfn(argv[i]);
+				char *pfname;
 				csync_check_usefullness(realname, recursive);
-				csync_mark(realname, 0, 0);
+				pfname=strdup(prefixencode(realname));
+				csync_mark(pfname, 0, 0);
 
 				if ( recursive ) {
 					char *where_rec = "";
 
 					if ( !strcmp(realname, "/") )
-						asprintf(&where_rec, "or 1");
+						ASPRINTF(&where_rec, "or 1=1");
 					else
-						asprintf(&where_rec, "or (filename > '%s/' "
-							"and filename < '%s0')",
-							url_encode(realname), url_encode(realname));
+						ASPRINTF(&where_rec, "UNION ALL SELECT filename from file where filename > '%s/' "
+							"and filename < '%s0'",
+							url_encode(pfname), url_encode(pfname));
 
 					SQL_BEGIN("Adding dirty entries recursively",
 						"SELECT filename FROM file WHERE filename = '%s' %s",
-						url_encode(realname), where_rec)
+						url_encode(pfname), where_rec)
 					{
-						char *filename = strdup(url_encode(SQL_V[0]));
+						char *filename = strdup(url_decode(SQL_V(0)));
 						csync_mark(filename, 0, 0);
 						free(filename);
 					} SQL_END;
 				}
+				free(pfname);
 			}
 			break;
 
 		case MODE_FORCE:
 			for (i=optind; i < argc; i++) {
 				char *realname = getrealfn(argv[i]);
+				char *pfname = strdup(prefixencode(realname));
 				char *where_rec = "";
 
 				if ( recursive ) {
 					if ( !strcmp(realname, "/") )
-						asprintf(&where_rec, "or 1");
+						ASPRINTF(&where_rec, "or 1=1");
 					else
-						asprintf(&where_rec, "or (filename > '%s/' "
+						ASPRINTF(&where_rec, "or (filename > '%s/' "
 							"and filename < '%s0')",
-							url_encode(realname), url_encode(realname));
+							url_encode(pfname), url_encode(pfname));
 				}
 
 				SQL("Mark file as to be forced",
-					"UPDATE dirty SET force = 1 WHERE filename = '%s' %s",
-					url_encode(realname), where_rec);
+					"UPDATE dirty SET forced = 1 WHERE filename = '%s' %s",
+					url_encode(pfname), where_rec);
 
 				if ( recursive )
 					free(where_rec);
+				free(pfname);
 			}
 			break;
 
@@ -695,7 +825,7 @@ found_a_group:;
 			SQL_BEGIN("DB Dump - Hint",
 				"SELECT recursive, filename FROM hint ORDER BY filename")
 			{
-				printf("%s\t%s\n", SQL_V[0], url_decode(SQL_V[1]));
+				printf("%s\t%s\n", (char*)SQL_V(0), url_decode(SQL_V(1)));
 				retval = -1;
 			} SQL_END;
 			break;
@@ -705,8 +835,8 @@ found_a_group:;
 			SQL_BEGIN("DB Dump - File",
 				"SELECT checktxt, filename FROM file ORDER BY filename")
 			{
-				if (csync_find_next(0, url_decode(SQL_V[1]))) {
-					printf("%s\t%s\n", url_decode(SQL_V[0]), url_decode(SQL_V[1]));
+				if (csync_find_next(0, url_decode(SQL_V(1)))) {
+					printf("%s\t%s\n", url_decode(SQL_V(0)), url_decode(SQL_V(1)));
 					retval = -1;
 				}
 			} SQL_END;
@@ -717,8 +847,8 @@ found_a_group:;
 			SQL_BEGIN("DB Dump - File",
 				"SELECT checktxt, filename FROM file ORDER BY filename")
 			{
-				if ( csync_match_file_host(url_decode(SQL_V[1]), argv[optind], argv[optind+1], 0) ) {
-					printf("%s\t%s\n", url_decode(SQL_V[0]), url_decode(SQL_V[1]));
+				if ( csync_match_file_host(url_decode(SQL_V(1)), argv[optind], argv[optind+1], 0) ) {
+					printf("%s\t%s\n", url_decode(SQL_V(0)), url_decode(SQL_V(1)));
 					retval = -1;
 				}
 			} SQL_END;
@@ -767,11 +897,11 @@ found_a_group:;
 		case MODE_LIST_DIRTY:
 			retval = 2;
 			SQL_BEGIN("DB Dump - Dirty",
-				"SELECT force, myname, peername, filename FROM dirty ORDER BY filename")
+				"SELECT forced, myname, peername, filename FROM dirty ORDER BY filename")
 			{
-				if (csync_find_next(0, url_decode(SQL_V[3]))) {
-					printf("%s\t%s\t%s\t%s\n", atoi(SQL_V[0]) ?  "force" : "chary",
-						url_decode(SQL_V[1]), url_decode(SQL_V[2]), url_decode(SQL_V[3]));
+				if (csync_find_next(0, url_decode(SQL_V(3)))) {
+					printf("%s\t%s\t%s\t%s\n", atoi(SQL_V(0)) ?  "force" : "chary",
+						url_decode(SQL_V(1)), url_decode(SQL_V(2)), url_decode(SQL_V(3)));
 					retval = -1;
 				}
 			} SQL_END;
@@ -787,11 +917,7 @@ found_a_group:;
 	csync_run_commands();
 	csync_db_close();
 
-	if ( csync_server_child_pid ) {
-		fprintf(stderr, "<%d> Connection closed.\n",
-				csync_server_child_pid);
-		fflush(stderr);
-	}
+	csync_debug(1, "Connection closed.\n");
 
 	if ( csync_error_count != 0 || (csync_messages_printed && csync_debug_level) )
 		csync_debug(0, "Finished with %d errors.\n", csync_error_count);
