@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,13 +15,17 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifndef CSYNC2_H
 #define CSYNC2_H 1
 
+#define CSYNC2_VERSION "2.0-rc1"
+
+#ifndef _GNU_SOURCE
 #define _GNU_SOURCE
+#endif
 
 #include "config.h"
 #include <stdio.h>
@@ -31,6 +35,24 @@
 #include <errno.h>
 
 
+#define DB_SCHEMA_VERSION 0
+
+/* asprintf with test for no memory */
+
+#define ASPRINTF(s, fmt, ...) do {\
+	int __ret = asprintf(s, fmt, ##__VA_ARGS__);\
+	if (__ret < 0) \
+		csync_fatal("Out of memory in asprintf at %s:%d\n", __FILE__, __LINE__);\
+} while (0)
+
+
+#define VASPRINTF(s, fmt, args) do {\
+	int __ret = vasprintf(s, fmt, args);\
+	if (__ret < 0) \
+		csync_fatal("Out of memory in vasprintf at %s:%d\n", __FILE__, __LINE__);\
+} while (0)
+
+
 /* action.c */
 
 extern void csync_schedule_commands(const char *filename, int islocal);
@@ -78,8 +100,67 @@ extern int conn_write(const void *buf, size_t count);
 
 extern void conn_printf(const char *fmt, ...);
 extern int conn_fgets(char *s, int size);
-extern int conn_gets(char *s, int size);
+extern size_t conn_gets(char *s, size_t size);
 
+/* Protocol encodes responses in plain text,
+ * everything starting with "OK (" is, well, OK.
+ *
+ * Anything not starting with that is an error response.
+ * In a few cases, behaviour differes depending on the
+ * exact error response string.
+ *
+ * Enum values are irrelevant for ABI (not communicated, only used internally),
+ * you may introduce more specific codes and string values, as long as you
+ * keep OK/ERR enums below/above CR_ERROR.
+ *
+ * See also: conn_response() and friends.
+ */
+enum connection_response {
+	CR_OK = 0,
+	/* OK response codes. They need all to start with "OK (" */
+	CR_OK_CMD_FINISHED,		/* "OK (cmd_finished).", */
+	CR_OK_DATA_FOLLOWS,		/* "OK (data_follows).", */
+	CR_OK_SEND_DATA,		/* "OK (send_data).", */
+	CR_OK_NOT_FOUND,		/* "OK (not_found).", */
+	CR_OK_PATH_NOT_FOUND,		/* "OK (path_not_found).", */
+	CR_OK_CU_LATER,			/* "OK (cu_later).", */
+	CR_OK_ACTIVATING_SSL,		/* "OK (activating_ssl).", */
+
+	CR_ERROR,
+	/* special error codes, MUST NOT start with "OK (" */
+	CR_ERR_CONN_CLOSED,
+	CR_ERR_ALSO_DIRTY_HERE,		/* "File is also marked dirty here!", */
+	CR_ERR_PARENT_DIR_MISSING,	/* "Parent dir missing.", */
+
+	CR_ERR_GROUP_LIST_ALREADY_SET,
+	CR_ERR_IDENTIFICATION_FAILED,
+	CR_ERR_PERM_DENIED_FOR_SLAVE,
+	CR_ERR_PERM_DENIED,
+	CR_ERR_SSL_EXPECTED,
+	CR_ERR_UNKNOWN_COMMAND,
+	CR_ERR_WIN32_EIO_CREATE_DIR,
+};
+
+static inline int is_ok_response(enum connection_response r)
+{
+	return r >= CR_OK && r < CR_ERROR;
+}
+
+/* converts to on-the-wire textual representation */
+extern const char *conn_response(enum connection_response);
+
+/* converts from on-the-wire textual representation */
+extern enum connection_response conn_response_to_enum(const char *);
+
+static inline void conn_resp(const int r)
+{
+	conn_printf("%s\n", conn_response(r));
+}
+
+static inline void conn_resp_zero(const int r)
+{
+	conn_printf("%s\n---\noctet-stream 0\n", conn_response(r));
+}
 
 /* db.c */
 
@@ -91,26 +172,56 @@ extern void* csync_db_begin(const char *err, const char *fmt, ...);
 extern int csync_db_next(void *vmx, const char *err,
 		int *pN, const char ***pazValue, const char ***pazColName);
 extern void csync_db_fin(void *vmx, const char *err);
+extern const void * csync_db_colblob(void *stmtx,int col);
+extern char *db_default_database(char *dbdir);
+
 
 #define SQL(e, s, ...) csync_db_sql(e, s, ##__VA_ARGS__)
 
+#if 0
+#if defined(HAVE_LIBSQLITE)
 #define SQL_BEGIN(e, s, ...) \
 { \
 	char *SQL_ERR = e; \
 	void *SQL_VM = csync_db_begin(SQL_ERR, s, ##__VA_ARGS__); \
 	int SQL_COUNT = 0; \
 	while (1) { \
-		const char **SQL_V, **SQL_N; \
+		const char **dataSQL_V, **dataSQL_N; \
 		int SQL_C; \
 		if ( !csync_db_next(SQL_VM, SQL_ERR, \
-					&SQL_C, &SQL_V, &SQL_N) ) break; \
+					&SQL_C, &dataSQL_V, &dataSQL_N) ) break; \
 		SQL_COUNT++;
 
+#define SQL_V(col) \
+	(dataSQL_V[(col)])
+#endif
+#endif
+
+// #if defined(HAVE_LIBSQLITE3)
+
+#define SQL_BEGIN(e, s, ...) \
+{ \
+	char *SQL_ERR = e; \
+	void *SQL_VM = csync_db_begin(SQL_ERR, s, ##__VA_ARGS__); \
+	int SQL_COUNT = 0; \
+\
+	if (SQL_VM) { \
+		while (1) { \
+			const char **dataSQL_V, **dataSQL_N; \
+			int SQL_C; \
+			if ( !csync_db_next(SQL_VM, SQL_ERR, \
+						&SQL_C, &dataSQL_V, &dataSQL_N) ) break; \
+			SQL_COUNT++;
+
+#define SQL_V(col) \
+	(csync_db_colblob(SQL_VM,(col)))
+// #endif
 #define SQL_FIN }{
 
 #define SQL_END \
+		} \
+		csync_db_fin(SQL_VM, SQL_ERR); \
 	} \
-	csync_db_fin(SQL_VM, SQL_ERR); \
 }
 
 extern int db_blocking_mode;
@@ -123,6 +234,8 @@ extern int csync_rs_check(const char *filename, int isreg);
 extern void csync_rs_sig(const char *filename);
 extern int csync_rs_delta(const char *filename);
 extern int csync_rs_patch(const char *filename);
+extern int mkpath(const char *path, mode_t mode);
+extern void split_dirname_basename(char *dirname, char* basename, const char *filepath);
 
 
 /* checktxt.c */
@@ -150,6 +263,7 @@ extern void csync_remove_old();
 /* daemon.c */
 
 extern void csync_daemon_session();
+extern int csync_copy_file(int fd_in, int fd_out);
 
 
 /* getrealfn.c */
@@ -170,6 +284,7 @@ const char *url_decode(const char *in);
 
 /* another ringbuffer here. so use it with care!! */
 const char *prefixsubst(const char *in);
+const char *prefixencode(const char *filename);
 
 
 /* textlist implementation */
@@ -233,12 +348,13 @@ struct csync_group_host {
 
 struct csync_group_pattern {
 	struct csync_group_pattern *next;
-	int isinclude, iscompare;
+	int isinclude, iscompare, star_matches_slashes;
 	const char *pattern;
 };
 
 struct csync_group_action_pattern {
 	struct csync_group_action_pattern *next;
+	int star_matches_slashes;
 	const char *pattern;
 };
 
@@ -253,6 +369,7 @@ struct csync_group_action {
 	struct csync_group_action_command *command;
 	const char *logfile;
 	int do_local;
+	int do_local_only;
 };
 
 struct csync_group {
@@ -301,8 +418,14 @@ extern struct csync_group  *csync_group;
 extern struct csync_prefix *csync_prefix;
 extern struct csync_nossl  *csync_nossl;
 
+extern unsigned csync_lock_timeout;
+extern char *csync_tempdir;
+
+extern char *csync_database;
+
 extern int csync_error_count;
 extern int csync_debug_level;
+extern int csync_syslog;
 extern FILE *csync_debug_out;
 
 extern long csync_last_printtime;
@@ -312,12 +435,14 @@ extern int csync_messages_printed;
 extern int csync_server_child_pid;
 extern int csync_timestamps;
 extern int csync_new_force;
-extern int csync_port;
 
 extern char myhostname[];
+extern int bind_to_myhostname;
+extern char *csync_port;
 extern char *active_grouplist;
 extern char *active_peerlist;
 
+extern char *systemdir;
 extern char *cfgname;
 
 extern int csync_ignore_uid;
@@ -328,7 +453,7 @@ extern int csync_dump_dir_fd;
 
 extern int csync_compare_mode;
 
-#ifdef HAVE_LIBGNUTLS_OPENSSL
+#ifdef HAVE_LIBGNUTLS
 extern int csync_conn_usessl;
 #endif
 
