@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
@@ -23,6 +23,7 @@
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
+#include <arpa/inet.h>
 #include <string.h>
 #include <fnmatch.h>
 #include <stdlib.h>
@@ -36,7 +37,10 @@
 #include <w32api/windows.h>
 #endif
 
-static char *cmd_error;
+static const char *cmd_error;
+
+int csync_setBackupFileStatus(char *filename, int backupDirLength);
+
 
 int csync_unlink(const char *filename, int ign)
 {
@@ -61,7 +65,7 @@ int csync_check_dirty(const char *filename, const char *peername, int isflush)
 		url_encode(filename))
 	{
 		rc = 1;
-		cmd_error = "File is also marked dirty here!";
+		cmd_error = conn_response(CR_ERR_ALSO_DIRTY_HERE);
 	} SQL_END;
 	if (rc && peername)
 		csync_mark(filename, peername, 0);
@@ -80,8 +84,13 @@ void csync_file_update(const char *filename, const char *peername)
 			url_encode(filename));
 	} else {
 		const char *checktxt = csync_genchecktxt(&st, filename, 0);
+
+		SQL("Deleting old record from file db",
+			"DELETE FROM file WHERE filename = '%s'",
+			url_encode(filename));
+
 		SQL("Insert record to file db",
-			"insert into file (filename, checktxt) values "
+			"INSERT INTO file (filename, checktxt) values "
 			"('%s', '%s')", url_encode(filename),
 			url_encode(checktxt));
 	}
@@ -94,77 +103,182 @@ void csync_file_flush(const char *filename)
 		url_encode(filename));
 }
 
-int csync_file_backup(const char *filename)
+int csync_file_backup(const char *filepath)
 {
 	static char error_buffer[1024];
 	const struct csync_group *g = NULL;
-	while ( (g=csync_find_next(g, filename)) ) {
-		if (g->backup_directory && g->backup_generations > 0) {
+	struct stat buf;
+	int rc;
+	/* ==============================================================================================
+	 * As of now, filepath may only contain prefixes but we may need to resolve other
+	 * dynamic references like environment variables, symbolic links, etc in future
+	 * if we plan to support those in later releases
+	 *==============================================================================================*/
+	const char *filename = prefixsubst(filepath);
+	int filename_len = strlen(filename);	//can filename be null?
+
+	while ((g = csync_find_next(g, filepath))) {
+		if (g->backup_directory && g->backup_generations > 1) {
 			int bak_dir_len = strlen(g->backup_directory);
-			int filename_len = strlen(filename);
 			char backup_filename[bak_dir_len + filename_len + 10];
 			char backup_otherfilename[bak_dir_len + filename_len + 10];
 			int fd_in, fd_out, i;
+			int lastSlash = 0;
+			mode_t mode;
 
-			fd_in = open(filename, O_RDONLY);
-			if (fd_in < 0) return 0;
+			csync_debug(1, "backup %s for group %s\n", filename, g->gname);
 
-			memcpy(backup_filename, g->backup_directory, bak_dir_len);
-			for (i=0; i<filename_len; i++)
-				backup_filename[bak_dir_len+i] =
-					filename[i] == '/' ? '_' : filename[i];
-			backup_filename[bak_dir_len] = '/';
-			memcpy(backup_otherfilename, backup_filename,
-					bak_dir_len + filename_len);
-
-			for (i=g->backup_generations-1; i; i--) {
-				snprintf(backup_filename+bak_dir_len+filename_len, 10, ".%d", i-1);
-				snprintf(backup_otherfilename+bak_dir_len+filename_len, 10, ".%d", i);
-				rename(backup_filename, backup_otherfilename);
+			if (stat(g->backup_directory, &buf) != 0) {
+				csync_debug(3, "backup directory configured is not present, so creating it");
+				if (mkpath(g->backup_directory, 0700)) {
+					csync_debug(1,
+						    "ERROR : unable to create backup directory %s ; backup failed \n",
+						    g->backup_directory);
+					return 1;
+				}
+			} else if (!S_ISDIR(buf.st_mode)) {
+				csync_debug(1,
+					    "ERROR : location configured for backup %s is not a directory; backup failed \n",
+					    g->backup_directory);
+				return 1;
+			}
+			// Skip generation of directories
+			rc = stat(filename, &buf);
+			if (S_ISDIR(buf.st_mode)) {
+				csync_debug(1, "directory. Skip generation \n");
+				return 0;
 			}
 
-			strcpy(backup_filename+bak_dir_len+filename_len, ".0");
-			fd_out = open(backup_filename, O_WRONLY|O_CREAT, 0600);
+			fd_in = open(filename, O_RDONLY);
+			if (fd_in < 0)
+				return 0;
+
+			memcpy(backup_filename, g->backup_directory, bak_dir_len);
+			backup_filename[bak_dir_len] = 0;
+			mode = 0700;
+
+			/* open coded strrchr?? why? */
+			for (i = filename_len; i > 0; i--)
+				if (filename[i] == '/') {
+					lastSlash = i;
+					break;
+				}
+
+			for (i = 0; i < filename_len; i++) {
+				// Create directories in filename
+				// TODO: Get the mode from the orig. dir
+				if (filename[i] == '/' && i <= lastSlash) {
+
+					backup_filename[bak_dir_len + i] = 0;
+
+					csync_debug(1, "mkdir %s \n", backup_filename);
+
+					mkdir(backup_filename, mode);
+					// Dont check the empty string.
+					if (i != 0)
+						csync_setBackupFileStatus(backup_filename, bak_dir_len);
+
+				}
+				backup_filename[bak_dir_len + i] = filename[i];
+			}
+
+			backup_filename[bak_dir_len + filename_len] = 0;
+			backup_filename[bak_dir_len] = '/';
+			memcpy(backup_otherfilename, backup_filename, bak_dir_len + filename_len);
+
+			for (i = g->backup_generations - 1; i; i--) {
+
+				if (i != 1)
+					snprintf(backup_filename + bak_dir_len + filename_len, 10, ".%d", i - 1);
+				backup_filename[bak_dir_len + filename_len] = '\0';
+				snprintf(backup_otherfilename + bak_dir_len + filename_len, 10, ".%d", i);
+
+				rc = rename(backup_filename, backup_otherfilename);
+				csync_debug(1,
+					    "renaming backup files '%s' to '%s'. rc = %d\n",
+					    backup_filename, backup_otherfilename, rc);
+			}
+
+			/* strcpy(backup_filename+bak_dir_len+filename_len, ""); */
+
+			fd_out = open(backup_filename, O_WRONLY | O_CREAT, 0600);
 
 			if (fd_out < 0) {
-				snprintf(error_buffer, 1024,
-						"Open error while backing up '%s': %s\n",
-						filename, strerror(errno));
+				snprintf(error_buffer, 1024, "Open error while backing up '%s': %s\n", filename, strerror(errno));
 				cmd_error = error_buffer;
 				close(fd_in);
 				return 1;
 			}
 
-			while (1) {
-				char buffer[512];
-				int read_len = read(fd_in, buffer, 512);
-				int write_len = 0;
+			csync_debug(1, "Copying data from %s to backup file %s \n", filename, backup_filename);
 
-				if (read_len <= 0)
-					break;
+			rc = csync_copy_file(fd_in, fd_out);
+			if (rc != 0) {
+				csync_debug(1, "csync_backup error 2\n");
 
-				while (write_len < read_len) {
-					int rc = write(fd_out, buffer+write_len, read_len-write_len);
-					if (rc <= 0) {
-						snprintf(error_buffer, 1024,
-								"Write error while backing up '%s': %s\n",
-								filename, strerror(errno));
-						cmd_error = error_buffer;
-						close(fd_in);
-						close(fd_out);
-						return 1;
-					}
-					write_len += rc;
-				}
+				snprintf(error_buffer, 1024, "Write error while backing up '%s': %s\n", filename, strerror(errno));
+
+				cmd_error = error_buffer;
+				// TODO verify file disapeared ? 
+				// 
+				// return 1;
 			}
-			close(fd_in);
-			close(fd_out);
+			csync_setBackupFileStatus(backup_filename, bak_dir_len);
 		}
 	}
-
 	return 0;
 }
 
+int csync_copy_file(int fd_in, int fd_out)
+{
+	char buffer[512];
+	int read_len = read(fd_in, buffer, 512);
+
+	while (read_len > 0) {
+		int write_len = 0;
+
+		while (write_len < read_len) {
+			int rc = write(fd_out, buffer + write_len,
+				       read_len - write_len);
+			if (rc == -1) {
+				close(fd_in);
+				close(fd_out);
+				//TODO verify return code. 
+				return errno;
+			}
+			write_len += rc;
+		}
+		read_len = read(fd_in, buffer, 512);
+	}
+	close(fd_in);
+	close(fd_out);
+	return 0;
+}
+
+/* get the mode from the orig directory.
+   Looking from the back_dir_len should produce the original dir.
+*/
+int csync_setBackupFileStatus(char *filename, int backupDirLength)
+{
+
+	struct stat buf;
+	int rc = stat((filename + backupDirLength), &buf);
+	if (rc == 0) {
+		csync_debug(0, "Stating original file %s rc: %d mode: %o", (filename + backupDirLength), rc, buf.st_mode);
+
+		rc = chown(filename, buf.st_uid, buf.st_gid);
+		csync_debug(0, "Changing owner of %s to user %d and group %d, rc= %d \n", filename, buf.st_uid, buf.st_gid, rc);
+
+		rc = chmod(filename, buf.st_mode);
+		csync_debug(0, "Changing mode of %s to mode %d, rc= %d \n", filename, buf.st_mode, rc);
+
+	} else {
+		csync_debug(0, "Error getting mode and owner ship from %s \n", (filename + backupDirLength));
+		return -1;
+	}
+	return 0;
+};
+
 struct csync_command {
 	char *text;
 	int check_perm;
@@ -210,49 +324,201 @@ struct csync_command cmdtab[] = {
 	{ 0,		0, 0, 0, 0, 0, 0	}
 };
 
+typedef union address {
+	struct sockaddr sa;
+	struct sockaddr_in sa_in;
+	struct sockaddr_in6 sa_in6;
+	struct sockaddr_storage ss;
+} address_t;
+
+const char *csync_inet_ntop(address_t *addr)
+{
+	char buf[INET6_ADDRSTRLEN];
+	sa_family_t af = addr->sa.sa_family;
+	return inet_ntop(af,
+		af == AF_INET  ? (void*)&addr->sa_in.sin_addr :
+		af == AF_INET6 ? (void*)&addr->sa_in6.sin6_addr : NULL,
+		buf, sizeof(buf));
+}
+
+/*
+ * Loops (to cater for multihomed peers) through the address list returned by
+ * gethostbyname(), returns 1 if any match with the address obtained from
+ * getpeername() during session startup.
+ * Otherwise returns 0 (-> identification failed).
+ *
+ * TODO switch to a getnameinfo in conn_open.
+ * TODO add a "pre-authenticated" pipe mode for use over ssh */
+int verify_peername(const char *name, address_t *peeraddr)
+{
+	sa_family_t af = peeraddr->sa.sa_family;
+	struct addrinfo hints;
+	struct addrinfo *result, *rp;
+	int try_mapped_ipv4;
+	int s;
+
+	/* Obtain address(es) matching host */
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_family = AF_UNSPEC;     /* Allow IPv4 or IPv6 */
+	hints.ai_socktype = SOCK_STREAM; /* Datagram socket */
+
+	s = getaddrinfo(name, NULL, &hints, &result);
+	if (s != 0) {
+		csync_debug(1, "getaddrinfo: %s\n", gai_strerror(s));
+		return 0;
+	}
+
+	try_mapped_ipv4 =
+		af == AF_INET6 &&
+		!memcmp(&peeraddr->sa_in6.sin6_addr,
+			"\0\0\0\0" "\0\0\0\0" "\0\0\xff\xff", 12);
+
+	/* getaddrinfo() returns a list of address structures.
+	 * Try each address. */
+
+	for (rp = result; rp != NULL; rp = rp->ai_next) {
+		/* both IPv4 */
+		if (af == AF_INET && rp->ai_family == AF_INET &&
+		    !memcmp(&((struct sockaddr_in*)rp->ai_addr)->sin_addr,
+			    &peeraddr->sa_in.sin_addr, sizeof(struct in_addr)))
+			break;
+		/* both IPv6 */
+		if (af == AF_INET6 && rp->ai_family == AF_INET6 &&
+		    !memcmp(&((struct sockaddr_in6*)rp->ai_addr)->sin6_addr,
+			    &peeraddr->sa_in6.sin6_addr, sizeof(struct in6_addr)))
+			break;
+		/* peeraddr IPv6, but actually ::ffff:I.P.v.4,
+		 * and forward lookup returned IPv4 only */
+		if (af == AF_INET6 && rp->ai_family == AF_INET &&
+		    try_mapped_ipv4 &&
+		    !memcmp(&((struct sockaddr_in*)rp->ai_addr)->sin_addr,
+			    (unsigned char*)&peeraddr->sa_in6.sin6_addr + 12,
+			    sizeof(struct in_addr)))
+			break;
+	}
+	freeaddrinfo(result);
+	if (rp != NULL) /* memcmp found a match */
+		return conn_check_peer_cert(name, 0);
+	return 0;
+}
+
+/* Why do all this fuzz, and not simply --assume-authenticated?
+ * To limit the impact of an accidental misconfiguration.
+ */
+void set_peername_from_env(address_t *p, const char *env)
+{
+	struct addrinfo hints = {
+		.ai_family = AF_UNSPEC,
+		.ai_socktype = SOCK_STREAM,
+		.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV,
+	};
+	struct addrinfo *result;
+	char *c;
+	int s;
+
+	char *val = getenv(env);
+	csync_debug(3, "getenv(%s): >>%s<<\n", env, val ?: "");
+	if (!val)
+		return;
+	val = strdup(val);
+	if (!val)
+		return;
+
+	c = strchr(val, ' ');
+	if (!c)
+		return;
+	*c = '\0';
+
+	s = getaddrinfo(val, NULL, &hints, &result);
+	if (s != 0) {
+		csync_debug(1, "getaddrinfo: %s\n", gai_strerror(s));
+		return;
+	}
+
+	/* getaddrinfo() may return a list of address structures.
+	 * Use the first one. */
+	if (result)
+		memcpy(p, result->ai_addr, result->ai_addrlen);
+	freeaddrinfo(result);
+}
+
+static int setup_tag(char *tag[32], char *line)
+{
+	int i = 0;
+	char *context;
+
+	tag[0] = strtok_r(line, "\t \r\n", &context);
+
+	while (tag[i] && i < 31) {
+		tag[++i] = strtok_r(NULL, "\t \r\n", &context);
+	}
+	while (i < 32) {
+		tag[i++] = "";
+	}
+
+	if (!tag[0][0]) {
+		return 0;
+	}
+
+	for (i = 0; i < 32; i++) {
+		tag[i] = strdup(url_decode(tag[i]));
+	}
+
+	return 1;
+}
+
+static void destroy_tag(char *tag[32])
+{
+	int i = 0;
+	for (i = 0; i < 32; i++)
+		free(tag[i]);
+}
 void csync_daemon_session()
 {
-	struct sockaddr_in peername;
-	struct hostent *hp;
-	int peerlen = sizeof(struct sockaddr_in);
-	char line[4096], *peer=0, *tag[32];
+	static char line[4 * 4096];
+	struct stat sb;
+	address_t peername = { .sa.sa_family = AF_UNSPEC, };
+	socklen_t peerlen = sizeof(peername);
+	char *peer=0, *tag[32];
 	int i;
 
-	if ( getpeername(0, (struct sockaddr*)&peername, &peerlen) == -1 )
-		csync_fatal("Can't run getpeername on fd 0: %s", strerror(errno));
 
-	while ( conn_gets(line, 4096) ) {
+	if (fstat(0, &sb))
+		csync_fatal("Can't run fstat on fd 0: %s", strerror(errno));
+
+	switch (sb.st_mode & S_IFMT) {
+	case S_IFSOCK:
+		if ( getpeername(0, &peername.sa, &peerlen) == -1 )
+			csync_fatal("Can't run getpeername on fd 0: %s", strerror(errno));
+		break;
+	case S_IFIFO:
+		set_peername_from_env(&peername, "SSH_CLIENT");
+		break;
+		/* fall through */
+	default:
+		csync_fatal("I'm only talking to sockets or pipes! %x\n", sb.st_mode & S_IFMT);
+		break;
+	}
+
+	while ( conn_gets(line, sizeof(line)) ) {
 		int cmdnr;
 
-		tag[i=0] = strtok(line, "\t \r\n");
-		while ( tag[i] && i < 31 )
-			tag[++i] = strtok(0, "\t \r\n");
-		while ( i < 32 )
-			tag[i++] = "";
-
-		if ( !tag[0][0] ) continue;
-
-		for (i=0; i<32; i++)
-			tag[i] = strdup(url_decode(tag[i]));
+		if (!setup_tag(tag, line))
+			continue;
 
 		for (cmdnr=0; cmdtab[cmdnr].text; cmdnr++)
 			if ( !strcasecmp(cmdtab[cmdnr].text, tag[0]) ) break;
 
 		if ( !cmdtab[cmdnr].text ) {
-			cmd_error = "Unkown command!";
+			cmd_error = conn_response(CR_ERR_UNKNOWN_COMMAND);
 			goto abort_cmd;
 		}
 
 		cmd_error = 0;
 
 		if ( cmdtab[cmdnr].need_ident && !peer ) {
-			union {
-				in_addr_t addr;
-				unsigned char oct[4];
-			} tmp;
-			tmp.addr = peername.sin_addr.s_addr;
-			conn_printf("Dear %d.%d.%d.%d, please identify first.\n",
-					tmp.oct[0], tmp.oct[1], tmp.oct[2], tmp.oct[3]);
+			conn_printf("Dear %s, please identify first.\n",
+				    csync_inet_ntop(&peername) ?: "stranger");
 			goto next_cmd;
 		}
 
@@ -268,18 +534,19 @@ void csync_daemon_session()
 			if ( perm ) {
 				if ( perm == 2 ) {
 					csync_mark(tag[2], peer, 0);
-					cmd_error = "Permission denied for slave!";
+					cmd_error = conn_response(CR_ERR_PERM_DENIED_FOR_SLAVE);
 				} else
-					cmd_error = "Permission denied!";
+					cmd_error = conn_response(CR_ERR_PERM_DENIED);
 				goto abort_cmd;
 			}
 		}
 
-		if ( cmdtab[cmdnr].check_dirty && csync_check_dirty(tag[2], peer,
-				cmdtab[cmdnr].action == A_FLUSH) ) goto abort_cmd;
+		if ( cmdtab[cmdnr].check_dirty &&
+			csync_check_dirty(tag[2], peer, cmdtab[cmdnr].action == A_FLUSH) )
+			goto abort_cmd;
 
 		if ( cmdtab[cmdnr].unlink )
-				csync_unlink(tag[2], cmdtab[cmdnr].unlink);
+			csync_unlink(tag[2], cmdtab[cmdnr].unlink);
 
 		switch ( cmdtab[cmdnr].action )
 		{
@@ -288,17 +555,23 @@ void csync_daemon_session()
 				struct stat st;
 
 				if ( lstat_strict(prefixsubst(tag[2]), &st) != 0 ) {
-					if ( errno == ENOENT )
-						conn_printf("OK (not_found).\n---\noctet-stream 0\n");
-					else
+					if ( errno == ENOENT ) {
+						struct stat sb;
+						char parent_dirname[strlen(tag[2])];
+						split_dirname_basename(parent_dirname, NULL, tag[2]);
+						if ( lstat_strict(prefixsubst(parent_dirname), &sb) != 0 )
+							cmd_error = conn_response(CR_ERR_PARENT_DIR_MISSING);
+						else
+							conn_resp_zero(CR_OK_PATH_NOT_FOUND);
+					} else
 						cmd_error = strerror(errno);
 					break;
 				} else
 					if ( csync_check_pure(tag[2]) ) {
-						conn_printf("OK (not_found).\n---\noctet-stream 0\n");
+						conn_resp_zero(CR_OK_NOT_FOUND);
 						break;
 					}
-				conn_printf("OK (data_follows).\n");
+				conn_resp(CR_OK_DATA_FOLLOWS);
 				conn_printf("%s\n", csync_genchecktxt(&st, tag[2], 1));
 
 				if ( S_ISREG(st.st_mode) )
@@ -321,7 +594,7 @@ void csync_daemon_session()
 					char buffer[512];
 					size_t rc;
 
-					conn_printf("OK (data_follows).\n");
+					conn_resp(CR_OK_DATA_FOLLOWS);
 					while ( (rc=fread(buffer, 1, 512, f)) > 0 )
 						if ( conn_write(buffer, rc) != rc ) {
 							conn_printf("[[ %s ]]", strerror(errno));
@@ -337,7 +610,7 @@ void csync_daemon_session()
 		case A_GETSZ:
 			{
 				struct stat sbuf;
-				conn_printf("OK (data_follows).\n");
+				conn_resp(CR_OK_DATA_FOLLOWS);
 				if (!lstat_strict(prefixsubst(tag[2]), &sbuf))
 					conn_printf("%ld\n", cmdtab[cmdnr].action == A_GETTM ?
 							(long)sbuf.st_mtime : (long)sbuf.st_size);
@@ -357,7 +630,7 @@ void csync_daemon_session()
 			break;
 		case A_PATCH:
 			if (!csync_file_backup(tag[2])) {
-				conn_printf("OK (send_data).\n");
+				conn_resp(CR_OK_SEND_DATA);
 				csync_rs_sig(tag[2]);
 				if (csync_rs_patch(tag[2]))
 					cmd_error = strerror(errno);
@@ -382,7 +655,7 @@ void csync_daemon_session()
 					if ( lstat_strict(prefixsubst(tag[2]), &st) != 0 || !S_ISDIR(st.st_mode)) {
 						csync_debug(1, "Win32 I/O Error %d in mkdir command: %s\n",
 								(int)GetLastError(), winfilename);
-						cmd_error = "Win32 I/O Error on CreateDirectory()";
+						cmd_error = conn_response(CR_ERR_WIN32_EIO_CREATE_DIR);
 					}
 				}
 			}
@@ -443,8 +716,8 @@ void csync_daemon_session()
 					strcmp(tag[2], "-") ? url_encode(tag[2]) : "",
 					strcmp(tag[2], "-") ? "'" : "")
 			{
-				if ( csync_match_file_host(url_decode(SQL_V[1]), tag[1], peer, (const char **)&tag[3]) )
-					conn_printf("%s\t%s\n", SQL_V[0], SQL_V[1]);
+				if ( csync_match_file_host(url_decode(SQL_V(1)), tag[1], peer, (const char **)&tag[3]) )
+					conn_printf("%s\t%s\n", SQL_V(0), SQL_V(1));
 			} SQL_END;
 			break;
 
@@ -454,18 +727,18 @@ void csync_daemon_session()
 				csync_debug_level = atoi(tag[1]);
 			break;
 		case A_HELLO:
-			if (peer) free(peer);
-			hp = gethostbyname(tag[1]);
-			if ( hp != 0 && peername.sin_family == hp->h_addrtype &&
-			     !memcmp(hp->h_addr, &peername.sin_addr, hp->h_length) &&
-			     conn_check_peer_cert(tag[1], 0)) {
+			if (peer) {
+				free(peer);
+				peer = NULL;
+			}
+			if (verify_peername(tag[1], &peername)) {
 				peer = strdup(tag[1]);
 			} else {
-				peer = 0;
-				cmd_error = "Identification failed!";
+				peer = NULL;
+				cmd_error = conn_response(CR_ERR_IDENTIFICATION_FAILED);
 				break;
 			}
-#ifdef HAVE_LIBGNUTLS_OPENSSL
+#ifdef HAVE_LIBGNUTLS
 			if (!csync_conn_usessl) {
 				struct csync_nossl *t;
 				for (t = csync_nossl; t; t=t->next) {
@@ -473,14 +746,14 @@ void csync_daemon_session()
 					     !fnmatch(t->pattern_to, peer, 0) )
 						goto conn_without_ssl_ok;
 				}
-				cmd_error = "SSL encrypted connection expected!";
+				cmd_error = conn_response(CR_ERR_SSL_EXPECTED);
 			}
 conn_without_ssl_ok:;
 #endif
 			break;
 		case A_GROUP:
 			if (active_grouplist) {
-				cmd_error = "Group list already set!";
+				cmd_error = conn_response(CR_ERR_GROUP_LIST_ALREADY_SET);
 			} else {
 				const struct csync_group *g;
 				int i, gnamelen;
@@ -507,8 +780,8 @@ found_asactive: ;
 			break;
 		case A_BYE:
 			for (i=0; i<32; i++)
-				tag[i] = strdup(url_decode(tag[i]));
-			conn_printf("OK (cu_later).\n");
+				free(tag[i]);
+			conn_resp(CR_OK_CU_LATER);
 			return;
 		}
 
@@ -525,11 +798,9 @@ abort_cmd:
 		if ( cmd_error )
 			conn_printf("%s\n", cmd_error);
 		else
-			conn_printf("OK (cmd_finished).\n");
+			conn_resp(CR_OK_CMD_FINISHED);
 
 next_cmd:
-		for (i=0; i<32; i++)
-			tag[i] = strdup(url_decode(tag[i]));
+		destroy_tag(tag);
 	}
 }
-
