@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,17 +15,17 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
-#include <sqlite.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <signal.h>
 #include <time.h>
+#include "db_api.h"
 
 #define DEADLOCK_MESSAGE \
 	"Database backend is exceedingly busy => Terminating (requesting retry).\n"
@@ -33,14 +33,16 @@
 int db_blocking_mode = 1;
 int db_sync_mode = 1;
 
-static sqlite *db = 0;
+extern int db_type; 
+static db_conn_p db = 0;
+// TODO make configurable
+int wait = 1;
 
 static int get_dblock_timeout()
 {
-	return getpid() % 7 + 12;
+	return getpid() % 7 + csync_lock_timeout;
 }
 
-
 static int tqueries_counter = -50;
 static time_t transaction_begin = 0;
 static time_t last_wait_cycle = 0;
@@ -58,7 +60,7 @@ void csync_db_alarmhandler(int signum)
 	begin_commit_recursion++;
 
 	csync_debug(2, "Database idle in transaction. Forcing COMMIT.\n");
-	SQL("COMMIT TRANSACTION", "COMMIT TRANSACTION");
+	SQL("COMMIT ", "COMMIT ");
 	tqueries_counter = -10;
 
 	begin_commit_recursion--;
@@ -82,7 +84,7 @@ void csync_db_maybegin()
 		transaction_begin = time(0);
 		if (!last_wait_cycle)
 			last_wait_cycle = transaction_begin;
-		SQL("BEGIN TRANSACTION", "BEGIN TRANSACTION");
+		SQL("BEGIN ", "BEGIN ");
 	}
 
 	begin_commit_recursion--;
@@ -103,9 +105,11 @@ void csync_db_maycommit()
 	now = time(0);
 
 	if ((now - last_wait_cycle) > 10) {
-		SQL("COMMIT TRANSACTION", "COMMIT TRANSACTION");
-		csync_debug(2, "Waiting 2 secs so others can lock the database (%d - %d)...\n", (int)now, (int)last_wait_cycle);
-		sleep(2);
+		SQL("COMMIT", "COMMIT ");
+		if (wait) {
+		  csync_debug(2, "Waiting %d secs so others can lock the database (%d - %d)...\n", wait, (int)now, (int)last_wait_cycle);
+		  sleep(wait);
+		}
 		last_wait_cycle = 0;
 		tqueries_counter = -10;
 		begin_commit_recursion--;
@@ -113,7 +117,7 @@ void csync_db_maycommit()
 	}
 
 	if ((tqueries_counter > 1000) || ((now - transaction_begin) > 3)) {
-		SQL("COMMIT TRANSACTION", "COMMIT TRANSACTION");
+	        SQL("COMMIT ", "COMMIT ");
 		tqueries_counter = 0;
 		begin_commit_recursion--;
 		return;
@@ -128,45 +132,23 @@ void csync_db_maycommit()
 
 void csync_db_open(const char *file)
 {
-	db = sqlite_open(file, 0, 0);
-	if ( db == 0 )
+        int rc = db_open(file, db_type, &db);
+	if ( rc != DB_OK )
 		csync_fatal("Can't open database: %s\n", file);
 
+	db_set_logger(db, csync_debug);
+
 	/* ignore errors on table creation */
 	in_sql_query++;
-	sqlite_exec(db,
-		"CREATE TABLE file ("
-		"	filename, checktxt,"
-		"	UNIQUE ( filename ) ON CONFLICT REPLACE"
-		")",
-		0, 0, 0);
-	sqlite_exec(db,
-		"CREATE TABLE dirty ("
-		"	filename, force, myname, peername,"
-		"	UNIQUE ( filename, peername ) ON CONFLICT IGNORE"
-		")",
-		0, 0, 0);
-	sqlite_exec(db,
-		"CREATE TABLE hint ("
-		"	filename, recursive,"
-		"	UNIQUE ( filename, recursive ) ON CONFLICT IGNORE"
-		")",
-		0, 0, 0);
-	sqlite_exec(db,
-		"CREATE TABLE action ("
-		"	filename, command, logfile,"
-		"	UNIQUE ( filename, command ) ON CONFLICT IGNORE"
-		")",
-		0, 0, 0);
-	sqlite_exec(db,
-		"CREATE TABLE x509_cert ("
-		"	peername, certdata,"
-		"	UNIQUE ( peername ) ON CONFLICT IGNORE"
-		")",
-		0, 0, 0);
+
+	if (db_schema_version(db) < DB_SCHEMA_VERSION)
+		if (db_upgrade_to_schema(db, DB_SCHEMA_VERSION) != DB_OK)
+			csync_fatal("Cannot create database tables (version requested = %d): %s\n", DB_SCHEMA_VERSION, db_errmsg(db));
+
 	if (!db_sync_mode)
-		sqlite_exec(db, "PRAGMA synchronous = OFF", 0, 0, 0);
+		db_exec(db, "PRAGMA synchronous = OFF");
 	in_sql_query--;
+	// return db;
 }
 
 void csync_db_close()
@@ -175,10 +157,10 @@ void csync_db_close()
 
 	begin_commit_recursion++;
 	if (tqueries_counter > 0) {
-		SQL("COMMIT TRANSACTION", "COMMIT TRANSACTION");
+	        SQL("COMMIT ", "COMMIT ");
 		tqueries_counter = -10;
 	}
-	sqlite_close(db);
+	db_close(db);
 	begin_commit_recursion--;
 	db = 0;
 }
@@ -190,7 +172,7 @@ void csync_db_sql(const char *err, const char *fmt, ...)
 	int rc, busyc = 0;
 
 	va_start(ap, fmt);
-	vasprintf(&sql, fmt, ap);
+	VASPRINTF(&sql, fmt, ap);
 	va_end(ap);
 
 	in_sql_query++;
@@ -199,15 +181,15 @@ void csync_db_sql(const char *err, const char *fmt, ...)
 	csync_debug(2, "SQL: %s\n", sql);
 
 	while (1) {
-		rc = sqlite_exec(db, sql, 0, 0, 0);
-		if ( rc != SQLITE_BUSY ) break;
-		if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
-		csync_debug(2, "Database is busy, sleeping a sec.\n");
-		sleep(1);
+	  rc = db_exec(db, sql);
+	  if ( rc != DB_BUSY ) break;
+	  if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
+	  csync_debug(2, "Database is busy, sleeping a sec.\n");
+	  sleep(1);
 	}
 
-	if ( rc != SQLITE_OK && err )
-		csync_fatal("Database Error: %s [%d]: %s\n", err, rc, sql);
+	if ( rc != DB_OK && err )
+		csync_fatal("Database Error: %s [%d]: %s on executing %s\n", err, rc, db_errmsg(db), sql);
 	free(sql);
 
 	csync_db_maycommit();
@@ -216,77 +198,122 @@ void csync_db_sql(const char *err, const char *fmt, ...)
 
 void* csync_db_begin(const char *err, const char *fmt, ...)
 {
-	sqlite_vm *vm;
+	db_stmt_p stmt = NULL;
 	char *sql;
 	va_list ap;
 	int rc, busyc = 0;
-
+	char *ppTail; 
 	va_start(ap, fmt);
-	vasprintf(&sql, fmt, ap);
+	VASPRINTF(&sql, fmt, ap);
 	va_end(ap);
 
 	in_sql_query++;
 	csync_db_maybegin();
 
 	csync_debug(2, "SQL: %s\n", sql);
-
 	while (1) {
-		rc = sqlite_compile(db, sql, 0, &vm, 0);
-		if ( rc != SQLITE_BUSY ) break;
+	        rc = db_prepare_stmt(db, sql, &stmt, &ppTail);
+		if ( rc != DB_BUSY ) break;
 		if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
 		csync_debug(2, "Database is busy, sleeping a sec.\n");
 		sleep(1);
 	}
 
-	if ( rc != SQLITE_OK && err )
-		csync_fatal("Database Error: %s [%d]: %s\n", err, rc, sql);
+	if ( rc != DB_OK && err )
+		csync_fatal("Database Error: %s [%d]: %s on executing %s\n", err, rc, db_errmsg(db), sql);
 	free(sql);
 
-	return vm;
+	return stmt;
+}
+
+const char *csync_db_get_column_text(void  *stmt, int column) {
+	return db_stmt_get_column_text(stmt, column);
+}
+
+int csync_db_get_column_int(void *stmt, int column) {
+  return db_stmt_get_column_int((db_stmt_p) stmt, column);
 }
 
 int csync_db_next(void *vmx, const char *err,
 		int *pN, const char ***pazValue, const char ***pazColName)
 {
-	sqlite_vm *vm = vmx;
+	db_stmt_p stmt = vmx;
 	int rc, busyc = 0;
 
 	csync_debug(4, "Trying to fetch a row from the database.\n");
 
 	while (1) {
-		rc = sqlite_step(vm, pN, pazValue, pazColName);
-		if ( rc != SQLITE_BUSY ) break;
-		if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
+		rc = db_stmt_next(stmt);
+		if ( rc != DB_BUSY ) 
+		  break;
+		if (busyc++ > get_dblock_timeout()) { 
+		  db = 0; 
+		  csync_fatal(DEADLOCK_MESSAGE); 
+		}
 		csync_debug(2, "Database is busy, sleeping a sec.\n");
 		sleep(1);
 	}
 
-	if ( rc != SQLITE_OK && rc != SQLITE_ROW &&
-	     rc != SQLITE_DONE && err )
-		csync_fatal("Database Error: %s [%d].\n", err, rc);
+	if ( rc != DB_OK && rc != DB_ROW &&
+	     rc != DB_DONE && err )
+		csync_fatal("Database Error: %s [%d]: %s\n", err, rc, db_errmsg(db));
 
-	return rc == SQLITE_ROW;
+	return rc == DB_ROW;
+}
+
+const void * csync_db_colblob(void *stmtx, int col) {
+       db_stmt_p stmt = stmtx;
+       const void *ptr = stmt->get_column_blob(stmt, col);
+       if (stmt->db && stmt->db->logger) {
+	 stmt->db->logger(4, "DB get blob: %s ", (char *) ptr);
+       }
+       return ptr;
 }
 
 void csync_db_fin(void *vmx, const char *err)
 {
-	sqlite_vm *vm = vmx;
+        db_stmt_p stmt = (db_stmt_p) vmx;
 	int rc, busyc = 0;
 
+	if (vmx == NULL)
+	   return;
+
 	csync_debug(2, "SQL Query finished.\n");
 
 	while (1) {
-		rc = sqlite_finalize(vm, 0);
-		if ( rc != SQLITE_BUSY ) break;
-		if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
-		csync_debug(2, "Database is busy, sleeping a sec.\n");
-		sleep(1);
+	  rc = db_stmt_close(stmt);
+	  if ( rc != DB_BUSY ) 
+	    break;
+	  if (busyc++ > get_dblock_timeout()) { db = 0; csync_fatal(DEADLOCK_MESSAGE); }
+	  csync_debug(2, "Database is busy, sleeping a sec.\n");
+	  sleep(1);
 	}
 
-	if ( rc != SQLITE_OK && err )
-		csync_fatal("Database Error: %s [%d].\n", err, rc);
+	if ( rc != DB_OK && err )
+		csync_fatal("Database Error: %s [%d]: %s\n", err, rc, db_errmsg(db));
 
 	csync_db_maycommit();
 	in_sql_query--;
 }
 
+char *db_default_database(char *dbdir)
+{
+	char *db;
+
+	if (!dbdir || !dbdir[0])
+		dbdir = DBDIR;
+
+#if defined(HAVE_SQLITE3)
+	ASPRINTF(&db, "sqlite3://%s/%s%s%s.db3", dbdir, myhostname, cfgname[0] ? "_" : "", cfgname);
+#elif defined(HAVE_SQLITE)
+	ASPRINTF(&db, "sqlite2://%s/%s%s%s.db", dbdir, myhostname, cfgname[0] ? "_" : "", cfgname);
+#elif defined(HAVE_MYSQL)
+	ASPRINTF(&db, "mysql://root@localhost/csync2_%s%s%s", myhostname, cfgname[0] ? "_" : "", cfgname);
+#elif defined(HAVE_POSTGRES)
+	ASPRINTF(&db, "pgsql://root@localhost/csync2_%s%s%s", myhostname, cfgname[0] ? "_" : "", cfgname);
+#else
+#error "No database backend available. Please install either libpg, libmysqlclient or libsqlite, reconfigure and recompile"
+#endif
+
+	return db;
+}
