@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,13 +15,14 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <unistd.h>
 #include <assert.h>
 
@@ -53,8 +54,11 @@ char *getrealfn(const char *filename)
 	/* make the path absolute */
 	if ( *tempfn != '/' ) {
 		char *t2, *t1 = my_get_current_dir_name();
-		asprintf(&t2, "%s/%s", t1, tempfn);
-		free(t1); free(tempfn); tempfn = t2;
+
+		ASPRINTF(&t2, "%s/%s", t1, tempfn);
+		free(t1);
+		free(tempfn);
+		tempfn = t2;
 	}
 
 	/* remove leading slashes from tempfn */
@@ -104,16 +108,20 @@ char *getrealfn(const char *filename)
 
 	/* ok - this might be ugly, but who cares .. */
 	{
-		char *oldpwd = my_get_current_dir_name();
+		int cwdfd = open(".", O_RDONLY);
+		if (cwdfd < 0)
+			csync_fatal("Cannot open current directory!\n");
 		if ( !chdir(tempfn) ) {
 			char *t2, *t1 = my_get_current_dir_name();
 			if ( st_mark ) {
-				asprintf(&t2, "%s/%s", t1, st_mark+1);
+				ASPRINTF(&t2, "%s/%s", t1, st_mark+1);
 				free(tempfn); free(t1); tempfn = t2;
 			} else {
 				free(tempfn); tempfn = t1;
 			}
-			chdir(oldpwd);
+			if (fchdir(cwdfd))
+				csync_fatal("Cannot chdir back to previous directory!\n");
+			close(cwdfd);
 		} else
 			if ( st_mark ) *st_mark = '/';
 	}
