@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2005  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
@@ -46,7 +46,7 @@ const char *prefixsubst(const char *in)
 			ringbuff_counter = (ringbuff_counter+1) % RINGBUFF_LEN;
 			if (ringbuff[ringbuff_counter])
 				free(ringbuff[ringbuff_counter]);
-			asprintf(&ringbuff[ringbuff_counter], "%s%s", p->path, path);
+			ASPRINTF(&ringbuff[ringbuff_counter], "%s%s", p->path, path);
 			return ringbuff[ringbuff_counter];
 		}
 	}
@@ -56,3 +56,35 @@ const char *prefixsubst(const char *in)
 	return 0;
 }
 
+const char *prefixencode(const char *filename) {
+#if __CYGWIN__
+	if (!strcmp(filename, "/")) {
+		filename = "/cygdrive";
+	}
+#endif
+	struct csync_prefix *p = csync_prefix;
+
+	/*
+	 * Canonicalized paths will always contain /
+	 * Prefixsubsted paths will probably contain %
+	 */
+	if (*filename == '/')
+		while (p) {
+			if (p->path) {
+				int p_len = strlen(p->path);
+				int f_len = strlen(filename);
+
+				if (p_len <= f_len && !strncmp(p->path, filename, p_len) &&
+						(filename[p_len] == '/' || !filename[p_len])) {
+					ringbuff_counter = (ringbuff_counter+1) % RINGBUFF_LEN;
+					if (ringbuff[ringbuff_counter])
+						free(ringbuff[ringbuff_counter]);
+					ASPRINTF(&ringbuff[ringbuff_counter], "%%%s%%%s", p->name, filename+p_len);
+					return ringbuff[ringbuff_counter];
+				}
+			}
+			p = p->next;
+		}
+	return filename;
+}
+
