@@ -1,7 +1,7 @@
 /*
  *  csync2 - cluster synchronization tool, 2nd generation
- *  LINBIT Information Technologies GmbH <http://www.linbit.com>
- *  Copyright (C) 2004, 2005, 2006  Clifford Wolf <clifford@clifford.at>
+ *  Copyright (C) 2004 - 2013 LINBIT Information Technologies GmbH
+ *  http://www.linbit.com; see also AUTHORS
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "csync2.h"
@@ -33,20 +33,28 @@
 
 static int connection_closed_error = 1;
 
-int read_conn_status(const char *file, const char *host)
+enum connection_response read_conn_status(const char *file, const char *host)
 {
 	char line[4096];
-	if ( conn_gets(line, 4096) ) {
-		if ( !strncmp(line, "OK (", 4) ) return 0;
+	enum connection_response conn_status;
+
+	if ( conn_gets(line, sizeof(line)) ) {
+		conn_status = conn_response_to_enum(line);
+
+		if (!is_ok_response(conn_status))
+			csync_error_count++;
 	} else {
+		csync_error_count++;
 		connection_closed_error = 1;
-		strcpy(line, "Connection closed.\n");
+		sprintf(line, "%s\n", conn_response(CR_ERR_CONN_CLOSED));
+		conn_status = CR_ERR_CONN_CLOSED;
 	}
 	if ( file )
-		csync_debug(0, "While syncing file %s:\n", file);
-	csync_debug(0, "ERROR from peer %s: %s", host, line);
-	csync_error_count++;
-	return !strcmp(line, "File is also marked dirty here!") ? 1 : 2;
+		csync_debug(2, "While syncing file %s:\n", file);
+	else
+		file = "<no file>";
+	csync_debug(3, "response from peer(%s): %s [%u] <- %s", file, host, conn_status, line);
+	return conn_status;
 }
 
 int connect_to_host(const char *peername)
@@ -70,9 +78,9 @@ int connect_to_host(const char *peername)
 	if ( conn_open(peername) ) return -1;
 
 	if ( use_ssl ) {
-#if HAVE_LIBGNUTLS_OPENSSL
+#if HAVE_LIBGNUTLS
 		conn_printf("SSL\n");
-		if ( read_conn_status(0, peername) ) {
+		if ( read_conn_status(0, peername) != CR_OK_ACTIVATING_SSL ) {
 			csync_debug(1, "SSL command failed.\n");
 			conn_close();
 			return -1;
@@ -87,7 +95,7 @@ int connect_to_host(const char *peername)
 	}
 
 	conn_printf("CONFIG %s\n", url_encode(cfgname));
-	if ( read_conn_status(0, peername) ) {
+	if (!is_ok_response(read_conn_status(NULL, peername))) {
 		csync_debug(1, "Config command failed.\n");
 		conn_close();
 		return -1;
@@ -95,7 +103,7 @@ int connect_to_host(const char *peername)
 
 	if (active_grouplist) {
 		conn_printf("GROUP %s\n", url_encode(active_grouplist));
-		if ( read_conn_status(0, peername) ) {
+		if (!is_ok_response(read_conn_status(NULL, peername))) {
 			csync_debug(1, "Group command failed.\n");
 			conn_close();
 			return -1;
@@ -145,7 +153,8 @@ static int get_master_slave_status(const char *peername, const char *filename)
 void csync_update_file_del(const char *peername,
 		const char *filename, int force, int dry_run)
 {
-	int last_conn_status = 0, auto_resolve_run = 0;
+	enum connection_response last_conn_status;
+	int auto_resolve_run = 0;
 	const char *key = csync_key(peername, filename);
 
 	if ( !key ) {
@@ -162,19 +171,29 @@ auto_resolve_entry_point:
 			return;
 		}
 		conn_printf("FLUSH %s %s\n", url_encode(key), url_encode(filename));
-		if ( read_conn_status(filename, peername) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto got_error;
 	} else {
+		static char chk1[4 * 4096];
+		const char *chk2 = "---";
 		int i, found_diff = 0;
 		int rs_check_result;
-		const char *chk2 = "---";
-		char chk1[4096];
 
 		conn_printf("SIG %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( read_conn_status(filename, peername) ) goto got_error;
 
-		if ( !conn_gets(chk1, 4096) ) goto got_error;
+		last_conn_status = read_conn_status(filename, peername);
+		if (last_conn_status == CR_ERR_PARENT_DIR_MISSING) {
+			/* In this case, this is not really an error. */
+			--csync_error_count;
+			goto already_gone;
+		}
+
+		if (!is_ok_response(last_conn_status))
+			goto got_error;
+
+		if ( !conn_gets(chk1, sizeof(chk1)) ) goto got_error;
 		for (i=0; chk1[i] && chk1[i] != '\n' && chk2[i]; i++)
 			if ( chk1[i] != chk2[i] ) {
 				csync_debug(2, "File is different on peer (cktxt char #%d).\n", i);
@@ -190,13 +209,15 @@ auto_resolve_entry_point:
 			csync_debug(2, "File is different on peer (rsync sig).\n");
 			found_diff=1;
 		}
-		if ( read_conn_status(filename, peername) ) goto got_error;
+		if (!is_ok_response(read_conn_status(filename, peername))) goto got_error;
 
 		if ( !found_diff ) {
+already_gone:
 			csync_debug(1, "File is already up to date on peer.\n");
 			if ( dry_run ) {
 				printf("?S: %-15s %s\n", peername, filename);
-				return;
+				// DS Remove local dirty, even in dry run
+				// return;
 			}
 			goto skip_action;
 		}
@@ -208,7 +229,10 @@ auto_resolve_entry_point:
 
 	conn_printf("DEL %s %s\n",
 			url_encode(key), url_encode(filename));
-	if ( (last_conn_status=read_conn_status(filename, peername)) )
+	last_conn_status = read_conn_status(filename, peername);
+	/* FIXME be more specific?
+	 * (last_conn_status == CR_ERR_ALSO_DIRTY_HERE) ?? */
+	if (!is_ok_response(last_conn_status))
 		goto maybe_auto_resolve;
 
 skip_action:
@@ -223,8 +247,9 @@ skip_action:
 	return;
 
 maybe_auto_resolve:
-	if (!auto_resolve_run && last_conn_status == 2)
+	if (!auto_resolve_run && (last_conn_status == CR_ERR_ALSO_DIRTY_HERE ) )
 	{
+		csync_debug(0,"auto resolving... response = %d",last_conn_status);
 		if (get_master_slave_status(peername, filename)) {
 			csync_debug(0, "Auto-resolving conflict: Won 'master/slave' test.\n");
 			auto_resolve_run = 1;
@@ -269,16 +294,17 @@ got_error:
 	csync_debug(1, "File stays in dirty state. Try again later...\n");
 }
 
-void csync_update_file_mod(const char *peername,
+enum connection_response csync_update_file_mod(const char *peername,
 		const char *filename, int force, int dry_run)
 {
 	struct stat st;
-	int last_conn_status = 0, auto_resolve_run = 0;
+	enum connection_response last_conn_status = CR_ERROR;
+	int auto_resolve_run = 0;
 	const char *key = csync_key(peername, filename);
 
 	if ( !key ) {
 		csync_debug(2, "Skipping file update %s on %s - not in my groups.\n", filename, peername);
-		return;
+		return	CR_ERROR;
 	}
 
 auto_resolve_entry_point:
@@ -293,27 +319,33 @@ auto_resolve_entry_point:
 	if ( force ) {
 		if ( dry_run ) {
 			printf("!M: %-15s %s\n", peername, filename);
-			return;
+			return CR_OK;
 		}
 		conn_printf("FLUSH %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( read_conn_status(filename, peername) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto got_error;
 	} else {
+		static char chk1[4 * 4096];
+		const char *chk2;
 		int i, found_diff = 0;
 		int rs_check_result;
-		char chk1[4096];
-		const char *chk2;
 
 		conn_printf("SIG %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( read_conn_status(filename, peername) ) goto got_error;
+		last_conn_status = read_conn_status(filename, peername);
 
-		if ( !conn_gets(chk1, 4096) ) goto got_error;
+		if (!is_ok_response(last_conn_status)) {
+			csync_debug(3, "error from peer\n");
+			goto got_error;
+		}
+
+		if ( !conn_gets(chk1, sizeof(chk1)) ) goto got_error;
 		chk2 = csync_genchecktxt(&st, filename, 1);
 		for (i=0; chk1[i] && chk1[i] != '\n' && chk2[i]; i++)
 			if ( chk1[i] != chk2[i] ) {
-				csync_debug(2, "File is different on peer (cktxt char #%d).\n", i);
+				csync_debug(2, "File %s is different on peer (cktxt char #%d).\n",filename, i);
 				csync_debug(2, ">>> PEER:  %s>>> LOCAL: %s\n", chk1, chk2);
 				found_diff=1;
 				break;
@@ -326,56 +358,70 @@ auto_resolve_entry_point:
 			csync_debug(2, "File is different on peer (rsync sig).\n");
 			found_diff=1;
 		}
-		if ( read_conn_status(filename, peername) ) goto got_error;
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
+			goto got_error;
 
 		if ( !found_diff ) {
 			csync_debug(1, "File is already up to date on peer.\n");
 			if ( dry_run ) {
 				printf("?S: %-15s %s\n", peername, filename);
-				return;
+				// DS also skip on dry_run 
+				// return;
 			}
 			goto skip_action;
 		}
 		if ( dry_run ) {
 			printf("?M: %-15s %s\n", peername, filename);
-			return;
+			return last_conn_status;
 		}
 	}
 
 	if ( S_ISREG(st.st_mode) ) {
 		conn_printf("PATCH %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		/* FIXME be more specific?
+		 * (last_conn_status != CR_OK_SEND_DATA) ??
+		 * (last_conn_status == CR_ERR_ALSO_DIRTY_HERE) ?? */
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
+
 		if ( csync_rs_delta(filename) ) {
-			read_conn_status(filename, peername);
+			//why is the response ignored?
+			last_conn_status = read_conn_status(filename, peername);
 			goto got_error;
 		}
-		if ( read_conn_status(filename, peername) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto got_error;
 	} else
 	if ( S_ISDIR(st.st_mode) ) {
 		conn_printf("MKDIR %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
 	} else
 	if ( S_ISCHR(st.st_mode) ) {
 		conn_printf("MKCHR %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
 	} else
 	if ( S_ISBLK(st.st_mode) ) {
 		conn_printf("MKBLK %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
 	} else
 	if ( S_ISFIFO(st.st_mode) ) {
 		conn_printf("MKFIFO %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
 	} else
 	if ( S_ISLNK(st.st_mode) ) {
@@ -387,7 +433,8 @@ auto_resolve_entry_point:
 			conn_printf("MKLINK %s %s %s\n",
 					url_encode(key), url_encode(filename),
 					url_encode(target));
-			if ( (last_conn_status = read_conn_status(filename, peername)) )
+			last_conn_status = read_conn_status(filename, peername);
+			if (!is_ok_response(last_conn_status))
 				goto maybe_auto_resolve;
 		} else {
 			csync_debug(1, "File is a symlink but radlink() failed.\n", st.st_mode);
@@ -397,7 +444,8 @@ auto_resolve_entry_point:
 	if ( S_ISSOCK(st.st_mode) ) {
 		conn_printf("MKSOCK %s %s\n",
 				url_encode(key), url_encode(filename));
-		if ( (last_conn_status = read_conn_status(filename, peername)) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto maybe_auto_resolve;
 	} else {
 		csync_debug(1, "File type (mode=%o) is not supported.\n", st.st_mode);
@@ -407,13 +455,15 @@ auto_resolve_entry_point:
 	conn_printf("SETOWN %s %s %d %d\n",
 			url_encode(key), url_encode(filename),
 			st.st_uid, st.st_gid);
-	if ( read_conn_status(filename, peername) )
+	last_conn_status = read_conn_status(filename, peername);
+	if (!is_ok_response(last_conn_status))
 		goto got_error;
 
 	if ( !S_ISLNK(st.st_mode) ) {
 		conn_printf("SETMOD %s %s %d\n", url_encode(key),
 				url_encode(filename), st.st_mode);
-		if ( read_conn_status(filename, peername) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto got_error;
 	}
 
@@ -422,7 +472,8 @@ skip_action:
 		conn_printf("SETIME %s %s %Ld\n",
 				url_encode(key), url_encode(filename),
 				(long long)st.st_mtime);
-		if ( read_conn_status(filename, peername) )
+		last_conn_status = read_conn_status(filename, peername);
+		if (!is_ok_response(last_conn_status))
 			goto got_error;
 	}
 
@@ -434,10 +485,10 @@ skip_action:
 	if (auto_resolve_run)
 		csync_error_count--;
 
-	return;
+	return last_conn_status;
 
 maybe_auto_resolve:
-	if (!auto_resolve_run && last_conn_status == 2)
+	if (!auto_resolve_run && (last_conn_status == CR_ERR_ALSO_DIRTY_HERE))
 	{
 		if (get_master_slave_status(peername, filename)) {
 			csync_debug(0, "Auto-resolving conflict: Won 'master/slave' test.\n");
@@ -466,7 +517,8 @@ maybe_auto_resolve:
 			case CSYNC_AUTO_METHOD_BIGGER:
 			case CSYNC_AUTO_METHOD_SMALLER:
 				{
-					char buffer[1024], *type, *cmd;
+					static char buffer[4 * 4096];
+					char *type, *cmd;
 					long remotedata, localdata;
 					struct stat sbuf;
 
@@ -480,9 +532,11 @@ maybe_auto_resolve:
 					}
 
 					conn_printf("%s %s %s\n", cmd, url_encode(key), url_encode(filename));
-					if ( read_conn_status(filename, peername) ) goto got_error_in_autoresolve;
+					last_conn_status = read_conn_status(filename, peername);
+					if (!is_ok_response(last_conn_status))
+						goto got_error;
 
-					if ( !conn_gets(buffer, 4096) ) goto got_error_in_autoresolve;
+					if ( !conn_gets(buffer, sizeof(buffer)) ) goto got_error_in_autoresolve;
 					remotedata = atol(buffer);
 
 					if (remotedata == -1)
@@ -520,6 +574,7 @@ got_error:
 got_error_in_autoresolve:
 		csync_debug(0, "ERROR: Auto-resolving failed. Giving up.\n");
 	csync_debug(1, "File stays in dirty state. Try again later...\n");
+	return last_conn_status;
 }
 
 int compare_files(const char *filename, const char *pattern, int recursive)
@@ -533,37 +588,173 @@ int compare_files(const char *filename, const char *pattern, int recursive)
 	return 0;
 }
 
-void csync_update_host(const char *peername,
-		const char **patlist, int patnum, int recursive, int dry_run)
+struct textlist *csync_find_files_recursive(const char *filename)
 {
-	struct textlist *tl = 0, *t, *next_t;
-	struct textlist *tl_mod = 0, **last_tn=&tl;
-	char *current_name = 0;
-	struct stat st;
+	const char *ue_filename = url_encode(filename);
+	struct textlist *tl = NULL;
+	struct stat sb;
+
+	/* ASCII '/' (0x2f) and '0' (0x30) happen to be
+	 * adjacent symbols.  That makes everything strictly
+	 * sorting between file/ and file0 the subdir tree. */
+	SQL_BEGIN("Query File Table for missing files on the peer",
+		"SELECT filename FROM file WHERE filename = '%s' UNION ALL "
+		"SELECT filename FROM file WHERE filename > '%s/' and filename < '%s0' "
+		"ORDER BY filename;",
+		ue_filename,
+		ue_filename, ue_filename)
+	{
+		const char *fn = url_decode(SQL_V(0));
+		if (lstat_strict(prefixsubst(fn), &sb) == 0 && csync_check_pure(fn) == 0)
+			textlist_add(&tl, fn, 0);
+	} SQL_END;
+
+	return tl;
+}
+
+void csync_mark_tl(struct textlist *tl, const char *peername)
+{
+	struct textlist *t;
+	for (t = tl; t != NULL; t = t->next)
+		csync_mark(t->value, peername, NULL);
+}
+
+struct update_context {
+	char *current_name;
+	const char *peername;
+	const char *ue_peername;
+	const char **patlist;
+	int patnum;
+
+	int recursive;
+	int dry_run;
+
+	char *missing_parent_dir;
+};
+
+enum connection_response conn_hello(struct update_context *c, struct textlist *t)
+{
+	enum connection_response r = CR_OK;
+	if ( !c->current_name || strcmp(c->current_name, t->value2) ) {
+		csync_debug(3, "Dirty item %s %s %d\n", t->value, t->value2, t->intvalue); 
+		conn_printf("HELLO %s\n", url_encode(t->value2));
+		r = read_conn_status(t->value, c->peername);
+		if (!is_ok_response(r))
+			return r;
+		free(c->current_name);
+		c->current_name = strdup(t->value2);
+	}
+	return r;
+}
+
+enum connection_response csync_update_tl_mod(struct textlist *tl_mod, struct update_context *c)
+{
+	struct textlist *t;
+	char *skip_subtree;
+	size_t skip_subtree_len;
+	enum connection_response r = CR_OK;
+
+	/* If we had a missing_parent_dir before, skip ahead, to make sure we
+	 * make progress in case we have more than one missing subtree,
+	 * and some conflict or other error when trying to re-create them.
+	 */
+	skip_subtree = c->missing_parent_dir;
+	skip_subtree_len = skip_subtree ? strlen(skip_subtree) : 0;
+
+	for (t = tl_mod; t != 0; t = t->next) {
+		if (skip_subtree) {
+			int cmp = strncmp(skip_subtree, t->value, skip_subtree_len);
+			if (cmp > 0)
+				continue;
+			if (cmp == 0 && t->value[skip_subtree_len] == '/') {
+					csync_debug(3, "Skipped %s\n", t->value);
+					continue;
+			}
+			skip_subtree = NULL;
+			skip_subtree_len = 0;
+		}
+		r = conn_hello(c, t);
+		if (!is_ok_response(r))
+			return r;
+		if (!connection_closed_error)
+			r = csync_update_file_mod(c->peername, t->value, t->intvalue, c->dry_run);
+
+		if (r == CR_ERR_PARENT_DIR_MISSING) {
+			struct textlist *tl = NULL;
+			int filename_length = strlen(t->value);
+			char parent_dirname[filename_length];
+
+			split_dirname_basename(parent_dirname, NULL, t->value);
+			csync_debug(1, "missing parent dir : %s\n", parent_dirname);
+			if (c->missing_parent_dir
+			&& strcmp(c->missing_parent_dir, parent_dirname) == 0) {
+				/* Don't take action, if we had this same dir
+				 * missing on a previous iteration.
+				 * Instead skip ahead the todo list.
+				 */
+				skip_subtree = t->value;
+				skip_subtree_len = strlen(skip_subtree);
+
+				/* change to generic error, in case this was
+				 * the last item on the todo list */
+				r = CR_ERROR;
+			} else {
+				tl = csync_find_files_recursive(parent_dirname);
+				csync_mark_tl(tl, c->peername);
+				textlist_free(tl);
+				free(c->missing_parent_dir);
+				c->missing_parent_dir = strdup(parent_dirname);
+				break;
+			}
+		}
+	}
+	return r;
+}
+
+struct textlist *csync_find_dirty(struct update_context *c)
+{
+	struct textlist *tl = NULL;
 
 	SQL_BEGIN("Get files for host from dirty table",
-		"SELECT filename, myname, force FROM dirty WHERE peername = '%s' "
-		"ORDER by filename ASC", url_encode(peername))
+		"SELECT filename, myname, forced FROM dirty WHERE peername = '%s' "
+		"ORDER by filename ASC", c->ue_peername)
 	{
-		const char *filename = url_decode(SQL_V[0]);
-		int i, use_this = patnum == 0;
-		for (i=0; i<patnum && !use_this; i++)
-			if ( compare_files(filename, patlist[i], recursive) ) use_this = 1;
+		const char *filename = url_decode(SQL_V(0));
+		int use_this = (c->patnum == 0);
+		int i;
+		for (i=0; i < c->patnum && !use_this; i++)
+			if (compare_files(filename, c->patlist[i], c->recursive))
+				use_this = 1;
 		if (use_this)
-			textlist_add2(&tl, filename, url_decode(SQL_V[1]), atoi(SQL_V[2]));
+			textlist_add2(&tl, filename, url_decode(SQL_V(1)), atoi(SQL_V(2)));
 	} SQL_END;
 
+	return tl;
+}
+
+void csync_update_host_c(struct update_context *c)
+{
+	struct textlist *tl, *t, *next_t;
+	struct textlist *tl_mod, **last_tn;
+	enum connection_response r = CR_OK;
+	int saved_csync_error_count = csync_error_count;
+
+	tl = csync_find_dirty(c);
+	tl_mod = NULL;
+	last_tn=&tl;
+
 	/* just return if there are no files to update */
 	if ( !tl ) return;
 
-	if ( connect_to_host(peername) ) {
+	if ( connect_to_host(c->peername) ) {
 		csync_error_count++;
-		csync_debug(0, "ERROR: Connection to remote host failed.\n");
+		csync_debug(0, "ERROR: Connection to remote host `%s' failed.\n", c->peername);
 		csync_debug(1, "Host stays in dirty state. "
 				"Try again later...\n");
 		return;
 	}
 
+redo:
 	/*
 	 * The SQL statement above creates a linked list. Due to the
 	 * way the linked list is created, it has the reversed order
@@ -578,55 +769,68 @@ void csync_update_host(const char *peername,
 	 *
 	 */
 	for (t = tl; t != 0; t = next_t) {
+		struct stat st;
 		next_t = t->next;
 		if ( !lstat_strict(prefixsubst(t->value), &st) != 0 && !csync_check_pure(t->value)) {
 			*last_tn = next_t;
 			t->next = tl_mod;
 			tl_mod = t;
 		} else {
-			if ( !current_name || strcmp(current_name, t->value2) ) {
-				conn_printf("HELLO %s\n", url_encode(t->value2));
-				if ( read_conn_status(t->value, peername) )
-					goto ident_failed_1;
-				current_name = t->value2;
-			}
+			r = conn_hello(c, t);
+			if (!is_ok_response(r))
+				goto ident_failed_1;
+
 			if (!connection_closed_error)
-				csync_update_file_del(peername,
-						t->value, t->intvalue, dry_run);
+				csync_update_file_del(c->peername,
+						t->value, t->intvalue, c->dry_run);
 ident_failed_1:
 			last_tn=&(t->next);
 		}
 	}
 
-	for (t = tl_mod; t != 0; t = t->next) {
-		if ( !current_name || strcmp(current_name, t->value2) ) {
-			conn_printf("HELLO %s\n", url_encode(t->value2));
-			if ( read_conn_status(t->value, peername) )
-				goto ident_failed_2;
-			current_name = t->value2;
-		}
-		if (!connection_closed_error)
-			csync_update_file_mod(peername,
-					t->value, t->intvalue, dry_run);
-ident_failed_2:;
-	}
+	r = csync_update_tl_mod(tl_mod, c);
 
 	textlist_free(tl_mod);
 	textlist_free(tl);
 
+	if (r == CR_ERR_PARENT_DIR_MISSING) {
+		csync_error_count = saved_csync_error_count;
+		tl = csync_find_dirty(c);
+		tl_mod = NULL;
+		last_tn=&tl;
+		goto redo;
+	}
+
 	conn_printf("BYE\n");
-	read_conn_status(0, peername);
+	read_conn_status(0, c->peername);//why is response ignored?
 	conn_close();
 }
 
+void csync_update_host(const char *peername,
+		const char **patlist, int patnum, int recursive, int dry_run)
+{
+	struct update_context c = {
+		.current_name = NULL,
+		.peername = peername,
+		.ue_peername = strdup(url_encode(peername)),
+		.patlist = patlist,
+		.patnum = patnum,
+		.recursive = recursive,
+		.dry_run = dry_run,
+		.missing_parent_dir = NULL,
+	};
+	csync_update_host_c(&c);
+}
+
+
 void csync_update(const char ** patlist, int patnum, int recursive, int dry_run)
 {
 	struct textlist *tl = 0, *t;
 
 	SQL_BEGIN("Get hosts from dirty table",
-		"SELECT peername FROM dirty GROUP BY peername ORDER BY random()")
+		"SELECT peername FROM dirty GROUP BY peername")
 	{
-		textlist_add(&tl, url_decode(SQL_V[0]), 0);
+		textlist_add(&tl, url_decode(SQL_V(0)), 0);
 	} SQL_END;
 
 	for (t = tl; t != 0; t = t->next) {
@@ -672,15 +876,17 @@ found_host_check:
 
 	if ( connect_to_host(peername) ) {
 		csync_error_count++;
-		csync_debug(0, "ERROR: Connection to remote host failed.\n");
+		csync_debug(0, "ERROR: Connection to remote host `%s' failed.\n", peername);
 		return 0;
 	}
 
 	conn_printf("HELLO %s\n", myname);
-	if ( read_conn_status(0, peername) ) goto finish;
+	if (!is_ok_response(read_conn_status(NULL, peername)))
+		goto finish;
 
 	conn_printf("TYPE %s %s\n", g->key, filename);
-	if ( read_conn_status(0, peername) ) goto finish;
+	if (!is_ok_response(read_conn_status(NULL, peername)))
+		goto finish;
 
 	/* FIXME
 	 * verify type of file first!
@@ -702,8 +908,9 @@ found_host_check:
 	old_sigpipe_handler = signal(SIGPIPE, SIG_IGN);
 	p = popen(buffer, "w");
 
-	while ( (rc=conn_read(buffer, 512)) > 0 )
-		fwrite(buffer, rc, 1, p);
+	while ( (rc=conn_read(buffer, 512)) > 0
+		&& fwrite(buffer, rc, 1, p) == rc)
+		;
 
 	fclose(p);
 	signal(SIGPIPE, old_sigpipe_handler);
@@ -721,7 +928,7 @@ int csync_insynctest_readline(char **file, char **checktxt)
 	if (*checktxt) free(*checktxt);
 	*file = *checktxt = 0;
 
-	if ( !conn_gets(inbuf, 2048) ) return 1;
+	if ( !conn_gets(inbuf, sizeof(inbuf)) ) return 1;
 	if ( inbuf[0] != 'v' ) {
 		if ( !strncmp(inbuf, "OK (", 4) ) {
 			csync_debug(2, "End of query results: %s", inbuf);
@@ -774,12 +981,17 @@ found_host_check:
 
 	if ( connect_to_host(peername) ) {
 		csync_error_count++;
-		csync_debug(0, "ERROR: Connection to remote host failed.\n");
+		csync_debug(0, "ERROR: Connection to remote host `%s' failed.\n", peername);
 		return 0;
 	}
 
 	conn_printf("HELLO %s\n", myname);
-	read_conn_status(0, peername);
+	if (!is_ok_response(read_conn_status(0, peername))) {
+		csync_debug(0, "ERROR: remote host %s did not accept my identification.\n", peername);
+		csync_error_count++;
+		conn_close();
+		return 0;
+	}
 
 	conn_printf("LIST %s %s", peername, filename ? url_encode(filename) : "-");
 	for (g = csync_group; g; g = g->next) {
@@ -798,7 +1010,7 @@ found_host:
 			filename ? url_encode(filename) : "",
 			filename ? "'" : "")
 	{
-		char *l_file = strdup(url_decode(SQL_V[1])), *l_checktxt = strdup(url_decode(SQL_V[0]));
+		char *l_file = strdup(url_decode(SQL_V(1))), *l_checktxt = strdup(url_decode(SQL_V(0)));
 		if ( csync_match_file_host(l_file, myname, peername, 0) ) {
 			if ( remote_eof ) {
 got_remote_eof:
@@ -862,7 +1074,7 @@ got_remote_eof:
 	if (r_checktxt) free(r_checktxt);
 
 	conn_printf("BYE\n");
-	read_conn_status(0, peername);
+	read_conn_status(0, peername);//why is response ignored?
 	conn_close();
 
 	for (diff_ent=diff_list; diff_ent; diff_ent=diff_ent->next)
@@ -936,17 +1148,17 @@ void csync_remove_old()
 		const struct csync_group *g = 0;
 		const struct csync_group_host *h;
 
-		const char *filename = url_decode(SQL_V[0]); 
+		const char *filename = url_decode(SQL_V(0)); 
 
 		while ((g=csync_find_next(g, filename)) != 0) {
-			if (!strcmp(g->myname, SQL_V[1]))
+			if (!strcmp(g->myname, SQL_V(1)))
 				for (h = g->host; h; h = h->next) {
-					if (!strcmp(h->hostname, SQL_V[2]))
+					if (!strcmp(h->hostname, SQL_V(2)))
 						goto this_dirty_record_is_ok;
 				}
 		}
 
-		textlist_add2(&tl, SQL_V[0], SQL_V[2], 0);
+		textlist_add2(&tl, SQL_V(0), SQL_V(2), 0);
 
 this_dirty_record_is_ok:
 		;
@@ -962,8 +1174,8 @@ this_dirty_record_is_ok:
 	SQL_BEGIN("Query file DB",
 	          "SELECT filename FROM file")
 	{
-		if (!csync_find_next(0, url_decode(SQL_V[0])))
-			textlist_add(&tl, SQL_V[0], 0);
+		if (!csync_find_next(0, url_decode(SQL_V(0))))
+			textlist_add(&tl, SQL_V(0), 0);
 	} SQL_END;
 	for (t = tl; t != 0; t = t->next) {
 		csync_debug(1, "Removing %s from file db.\n", t->value);
